from mimeo import tools
from mimeo.resources.exc import ResourceNotFoundError
from tests.utils import assert_throws


def test_get_resource_existing():
    with tools.get_resource("logging.yaml") as resource:
        assert resource.name.endswith("resources/logging.yaml")


@assert_throws(err_type=ResourceNotFoundError,
               msg="No such resource: [{res}]",
               res="non-existing-file.yaml")
def test_get_resource_non_existing():
    tools.get_resource("non-existing-file.yaml")



"""The Mimeograph Exceptions module.

It contains all custom exceptions related to the highest Mimeograph level:
    * NotRunningMimeograph
        A custom Exception class for not running Mimeograph instance.
"""


from __future__ import annotations


class NotRunningMimeographError(Exception):
    """A custom Exception class for not running Mimeograph instance.

    Raised when using submit() method without running Mimeograph first.
    """

    def __init__(
            self,
    ):
        """Initialize NotRunningMimeograph exception with details.

        Extends Exception constructor with a constant message.
        """
        super().__init__("The Mimeograph instance is not running!")



"""The Mimeo File Consumer module.

It exports only one class:
    * FileConsumer
        A Consumer implementation saving data in the filesystem.
"""
from __future__ import annotations

import logging
from pathlib import Path
from typing import Collection, Generator

import aiofiles

from mimeo.config.mimeo_config import MimeoOutput
from mimeo.consumers import Consumer

logger = logging.getLogger(__name__)


class FileConsumer(Consumer):
    """A Consumer implementation saving data in the filesystem.

    This Consumer is instantiated for the 'file' output direction
    and saves data produced by Mimeo as files using Mimeo Output Details.

    Methods
    -------
    consume
        Save data generated by Mimeo into a file.

    Attributes
    ----------
    directory : str
        A directory path to save files within
    output_path_tmplt : str
        An output file path template
        (every file has its index inside the actual path)
    """

    def __init__(
            self,
            output: MimeoOutput,
    ):
        """Initialize FileConsumer class.

        Parameters
        ----------
        output : MimeoOutput
            Configured Mimeo Output Details
        """
        self.directory: str = output.directory_path
        self.output_path_tmplt: str = f"{self.directory}/{output.file_name}"

    async def consume(
            self,
            data: Collection | Generator,
    ) -> None:
        """Save data generated by Mimeo into a file.

        It is an implementation of Consumer's abstract method.
        If the output directory does not exist it is created.
        Every file name has an index inside its path.

        Parameters
        ----------
        data : Collection | Generator
            Stringified data generated by Mimeo
        """
        count = 0
        for data_unit in data:
            logger.fine("Consuming data [%s]", data_unit)
            if not Path(self.directory).exists():
                logger.info("Creating output directory [%s]", self.directory)
                Path(self.directory).mkdir(parents=True, exist_ok=True)

            count += 1
            file_name = self.output_path_tmplt.format(count)

            logger.info("Writing data into file [%s]", file_name)
            async with aiofiles.open(file_name, mode="w") as file:
                await file.write(data_unit)



"""The Mimeo Generators package.

It contains modules supporting the Mimeo Config output formats:
* generator
    The Mimeo Generator module.
* generator_factory
    The Mimeo Generator Factory module.
* json_generator
    The Mimeo JSON Generator module.
* xml_generator
    The Mimeo XML Generator module.
* exc
    The Mimeo Generators Exceptions module.

This package exports the following classes:
* Generator:
    An abstract class for data generators in Mimeo.
* GeneratorFactory:
    A Factory class instantiating a Generator based on Mimeo Config.
* JSONGenerator:
    A Generator implementation producing data in the JSON format.
    Corresponds to the 'json' output format
* XMLGenerator:
    A Generator implementation producing data in the XML format.
    Corresponds to the 'xml' output format

To use this package, simply import the desired class:
    from mimeo.generators import GeneratorFactory
    from mimeo.generators.exc import UnsupportedStructureError
"""
from .generator import Generator
from .json_generator import JSONGenerator
from .xml_generator import XMLGenerator
from .generator_factory import GeneratorFactory

__all__ = ["Generator", "JSONGenerator", "XMLGenerator", "GeneratorFactory"]



import shutil
import time
from pathlib import Path
from xml.etree import ElementTree as ElemTree

import pytest

from mimeo import Mimeograph
from mimeo.config import MimeoConfigFactory
from mimeo.exc import NotRunningMimeographError
from tests.utils import assert_throws


@pytest.fixture(autouse=True)
def _teardown():
    yield
    # Teardown
    shutil.rmtree("test_mimeograph-dir", ignore_errors=True)


def test_process_xml():
    config = {
        "output": {
            "direction": "file",
            "format": "xml",
            "indent": 4,
            "xml_declaration": True,
            "directory_path": "test_mimeograph-dir",
            "file_name": "output",
        },
        "_templates_": [
            {
                "count": 10,
                "model": {
                    "SomeEntity": {
                        "ChildNode1": 1,
                        "ChildNode2": "value-2",
                        "ChildNode3": True,
                    },
                },
            },
        ],
    }
    mimeo_config = MimeoConfigFactory.parse(config)
    assert not Path("test_mimeograph-dir").exists()
    Mimeograph.process(mimeo_config)
    assert Path("test_mimeograph-dir").exists()
    for i in range(1, 11):
        file_path = f"test_mimeograph-dir/output-{i}.xml"
        assert Path(file_path).exists()

        with Path(file_path).open() as file:
            assert file.readline() == '<?xml version="1.0" encoding="utf-8"?>\n'
            assert file.readline() == "<SomeEntity>\n"
            assert file.readline() == "    <ChildNode1>1</ChildNode1>\n"
            assert file.readline() == "    <ChildNode2>value-2</ChildNode2>\n"
            assert file.readline() == "    <ChildNode3>true</ChildNode3>\n"
            assert file.readline() == "</SomeEntity>\n"


def test_process_json():
    config = {
        "output": {
            "direction": "file",
            "format": "json",
            "indent": 4,
            "directory_path": "test_mimeograph-dir",
            "file_name": "output",
        },
        "_templates_": [
            {
                "count": 10,
                "model": {
                    "SomeEntity": {
                        "ChildNode1": 1,
                        "ChildNode2": "value-2",
                        "ChildNode3": True,
                    },
                },
            },
        ],
    }
    mimeo_config = MimeoConfigFactory.parse(config)
    assert not Path("test_mimeograph-dir").exists()
    Mimeograph.process(mimeo_config)
    assert Path("test_mimeograph-dir").exists()
    for i in range(1, 11):
        file_path = f"test_mimeograph-dir/output-{i}.json"
        assert Path(file_path).exists()

        with Path(file_path).open() as file:
            assert file.readline() == "{\n"
            assert file.readline() == '    "SomeEntity": {\n'
            assert file.readline() == '        "ChildNode1": 1,\n'
            assert file.readline() == '        "ChildNode2": "value-2",\n'
            assert file.readline() == '        "ChildNode3": true\n'
            assert file.readline() == "    }\n"
            assert file.readline() == "}"


def test_generate_xml():
    config = {
        "output": {
            "direction": "file",
            "format": "xml",
            "indent": 4,
            "xml_declaration": True,
            "directory_path": "test_mimeograph-dir",
            "file_name": "output",
        },
        "_templates_": [
            {
                "count": 10,
                "model": {
                    "SomeEntity": {
                        "ChildNode1": 1,
                        "ChildNode2": "value-2",
                        "ChildNode3": True,
                    },
                },
            },
        ],
    }
    mimeo_config = MimeoConfigFactory.parse(config)

    count = 0
    for data in Mimeograph.generate(mimeo_config):
        assert isinstance(data, ElemTree.Element)

        assert data.tag == "SomeEntity"
        assert data.attrib == {}
        assert len(list(data)) == 3  # number of children

        child = data.find("ChildNode1")
        assert child.tag == "ChildNode1"
        assert child.attrib == {}
        assert child.text == "1"
        assert len(list(child)) == 0  # number of children

        child = data.find("ChildNode2")
        assert child.tag == "ChildNode2"
        assert child.attrib == {}
        assert child.text == "value-2"
        assert len(list(child)) == 0  # number of children

        child = data.find("ChildNode3")
        assert child.tag == "ChildNode3"
        assert child.attrib == {}
        assert child.text == "true"
        assert len(list(child)) == 0  # number of children

        count += 1

    assert count == 10


def test_generate_xml_stringified():
    config = {
        "output": {
            "direction": "file",
            "format": "xml",
            "indent": 4,
            "xml_declaration": True,
            "directory_path": "test_mimeograph-dir",
            "file_name": "output",
        },
        "_templates_": [
            {
                "count": 10,
                "model": {
                    "SomeEntity": {
                        "ChildNode1": 1,
                        "ChildNode2": "value-2",
                        "ChildNode3": True,
                    },
                },
            },
        ],
    }
    mimeo_config = MimeoConfigFactory.parse(config)

    count = 0
    for data in Mimeograph.generate(mimeo_config, stringify=True):
        assert isinstance(data, str)

        assert data == ('<?xml version="1.0" encoding="utf-8"?>\n'
                        "<SomeEntity>\n"
                        "    <ChildNode1>1</ChildNode1>\n"
                        "    <ChildNode2>value-2</ChildNode2>\n"
                        "    <ChildNode3>true</ChildNode3>\n"
                        "</SomeEntity>\n")

        count += 1

    assert count == 10


def test_generate_json():
    config = {
        "output": {
            "direction": "file",
            "format": "json",
            "indent": 4,
            "xml_declaration": True,
            "directory_path": "test_mimeograph-dir",
            "file_name": "output",
        },
        "_templates_": [
            {
                "count": 10,
                "model": {
                    "SomeEntity": {
                        "ChildNode1": 1,
                        "ChildNode2": "value-2",
                        "ChildNode3": True,
                    },
                },
            },
        ],
    }
    mimeo_config = MimeoConfigFactory.parse(config)

    count = 0
    for data in Mimeograph.generate(mimeo_config):
        assert isinstance(data, dict)

        assert data == {
            "SomeEntity": {
                "ChildNode1": 1,
                "ChildNode2": "value-2",
                "ChildNode3": True,
            },
        }
        count += 1

    assert count == 10


def test_generate_json_stringified():
    config = {
        "output": {
            "direction": "file",
            "format": "json",
            "indent": 4,
            "directory_path": "test_mimeograph-dir",
            "file_name": "output",
        },
        "_templates_": [
            {
                "count": 10,
                "model": {
                    "SomeEntity": {
                        "ChildNode1": 1,
                        "ChildNode2": "value-2",
                        "ChildNode3": True,
                    },
                },
            },
        ],
    }
    mimeo_config = MimeoConfigFactory.parse(config)

    count = 0
    for data in Mimeograph.generate(mimeo_config, stringify=True):
        assert isinstance(data, str)

        assert data == ("{\n"
                        '    "SomeEntity": {\n'
                        '        "ChildNode1": 1,\n'
                        '        "ChildNode2": "value-2",\n'
                        '        "ChildNode3": true\n'
                        "    }\n"
                        "}")

        count += 1

    assert count == 10


def test_consume_xml():
    config = {
        "output": {
            "direction": "file",
            "format": "xml",
            "indent": 4,
            "xml_declaration": True,
            "directory_path": "test_mimeograph-dir",
            "file_name": "output",
        },
        "_templates_": [
            {
                "count": 10,
                "model": {
                    "SomeEntity": {
                        "ChildNode1": 1,
                        "ChildNode2": "value-2",
                        "ChildNode3": True,
                    },
                },
            },
        ],
    }
    mimeo_config = MimeoConfigFactory.parse(config)
    assert not Path("test_mimeograph-dir").exists()
    data = Mimeograph.generate(mimeo_config, stringify=True)
    Mimeograph.consume(mimeo_config, data)
    assert Path("test_mimeograph-dir").exists()
    for i in range(1, 11):
        file_path = f"test_mimeograph-dir/output-{i}.xml"
        assert Path(file_path).exists()

        with Path(file_path).open() as file:
            assert file.readline() == '<?xml version="1.0" encoding="utf-8"?>\n'
            assert file.readline() == "<SomeEntity>\n"
            assert file.readline() == "    <ChildNode1>1</ChildNode1>\n"
            assert file.readline() == "    <ChildNode2>value-2</ChildNode2>\n"
            assert file.readline() == "    <ChildNode3>true</ChildNode3>\n"
            assert file.readline() == "</SomeEntity>\n"


def test_consume_json():
    config = {
        "output": {
            "direction": "file",
            "format": "json",
            "indent": 4,
            "directory_path": "test_mimeograph-dir",
            "file_name": "output",
        },
        "_templates_": [
            {
                "count": 10,
                "model": {
                    "SomeEntity": {
                        "ChildNode1": 1,
                        "ChildNode2": "value-2",
                        "ChildNode3": True,
                    },
                },
            },
        ],
    }
    mimeo_config = MimeoConfigFactory.parse(config)
    assert not Path("test_mimeograph-dir").exists()
    data = Mimeograph.generate(mimeo_config, stringify=True)
    Mimeograph.consume(mimeo_config, data)
    assert Path("test_mimeograph-dir").exists()
    for i in range(1, 11):
        file_path = f"test_mimeograph-dir/output-{i}.json"
        assert Path(file_path).exists()

        with Path(file_path).open() as file:
            assert file.readline() == "{\n"
            assert file.readline() == '    "SomeEntity": {\n'
            assert file.readline() == '        "ChildNode1": 1,\n'
            assert file.readline() == '        "ChildNode2": "value-2",\n'
            assert file.readline() == '        "ChildNode3": true\n'
            assert file.readline() == "    }\n"
            assert file.readline() == "}"


def test_submit_using_run_and_stop():
    config = {
        "output": {
            "direction": "file",
            "format": "xml",
            "indent": 4,
            "xml_declaration": True,
            "directory_path": "test_mimeograph-dir",
            "file_name": "output",
        },
        "_templates_": [
            {
                "count": 10,
                "model": {
                    "SomeEntity": {
                        "ChildNode1": 1,
                        "ChildNode2": "value-2",
                        "ChildNode3": True,
                    },
                },
            },
        ],
    }
    mimeo_config = MimeoConfigFactory.parse(config)
    assert not Path("test_mimeograph-dir").exists()
    mimeo = Mimeograph()
    mimeo.run()
    mimeo.submit(("xml-config", mimeo_config))
    mimeo.stop()
    assert Path("test_mimeograph-dir").exists()
    for i in range(1, 11):
        file_path = f"test_mimeograph-dir/output-{i}.xml"
        assert Path(file_path).exists()

        with Path(file_path).open() as file:
            assert file.readline() == '<?xml version="1.0" encoding="utf-8"?>\n'
            assert file.readline() == "<SomeEntity>\n"
            assert file.readline() == "    <ChildNode1>1</ChildNode1>\n"
            assert file.readline() == "    <ChildNode2>value-2</ChildNode2>\n"
            assert file.readline() == "    <ChildNode3>true</ChildNode3>\n"
            assert file.readline() == "</SomeEntity>\n"


def test_submit_as_context_manager():
    config = {
        "output": {
            "direction": "file",
            "format": "json",
            "indent": 4,
            "directory_path": "test_mimeograph-dir",
            "file_name": "output",
        },
        "_templates_": [
            {
                "count": 10,
                "model": {
                    "SomeEntity": {
                        "ChildNode1": 1,
                        "ChildNode2": "value-2",
                        "ChildNode3": True,
                    },
                },
            },
        ],
    }
    mimeo_config = MimeoConfigFactory.parse(config)
    assert not Path("test_mimeograph-dir").exists()
    with Mimeograph() as mimeo:
        mimeo.submit(("json-config", mimeo_config))
    assert Path("test_mimeograph-dir").exists()
    for i in range(1, 11):
        file_path = f"test_mimeograph-dir/output-{i}.json"
        assert Path(file_path).exists()

        with Path(file_path).open() as file:
            assert file.readline() == "{\n"
            assert file.readline() == '    "SomeEntity": {\n'
            assert file.readline() == '        "ChildNode1": 1,\n'
            assert file.readline() == '        "ChildNode2": "value-2",\n'
            assert file.readline() == '        "ChildNode3": true\n'
            assert file.readline() == "    }\n"
            assert file.readline() == "}"


@assert_throws(err_type=NotRunningMimeographError,
               msg="The Mimeograph instance is not running!")
def test_submit_when_non_running():
    config = {
        "output": {
            "direction": "file",
            "format": "xml",
            "indent": 4,
            "directory_path": "test_mimeograph-dir",
            "file_name": "output",
        },
        "_templates_": [
            {
                "count": 10,
                "model": {
                    "SomeEntity": {
                        "ChildNode1": 1,
                        "ChildNode2": "value-2",
                        "ChildNode3": True,
                    },
                },
            },
        ],
    }
    mimeo_config = MimeoConfigFactory.parse(config)
    mimeo = Mimeograph()
    mimeo.submit(("xml-config", mimeo_config))


def test_default_number_of_workers():
    assert Mimeograph()._consumer_workers > 4


def test_custom_number_of_workers():
    assert Mimeograph(workers=1)._consumer_workers == 1


def test_config_failed_at_generation_step():
    config = {
        "output": {
            "direction": "file",
            "format": "xml",
            "directory_path": "test_mimeograph-dir",
            "file_name": "output",
        },
        "_templates_": [
            {
                "count": 10,
                "model": {
                    "SomeEntity": {
                        "ChildNode1": "Bolívar Soberano",  # Non-ascii character
                    },
                },
            },
        ],
    }
    mimeo_config = MimeoConfigFactory.parse(config)
    assert not Path("test_mimeograph-dir").exists()
    with Mimeograph() as mimeo:
        mimeo.submit(("non-ascii-character-config", mimeo_config))
        time.sleep(0.05)
        assert mimeo._failed_configs == ["non-ascii-character-config"]
    assert not Path("test_mimeograph-dir").exists()


def test_config_failed_at_consuming_step():
    config = {
        "output": {
            "direction": "http",
            "format": "xml",
            "host": "localhost",
            "port": 8080,
            "endpoint": "/documents",
            "username": "admin",
            "password": "admin",
        },
        "_templates_": [
            {
                "count": 10,
                "model": {
                    "SomeEntity": {
                        "ChildNode1": 1,
                    },
                },
            },
        ],
    }
    mimeo_config = MimeoConfigFactory.parse(config)
    assert not Path("test_mimeograph-dir").exists()
    with Mimeograph() as mimeo:
        mimeo.submit(("no-connection-config", mimeo_config))
        time.sleep(0.05)
        assert mimeo._failed_configs == ["no-connection-config"]
    assert not Path("test_mimeograph-dir").exists()



"""The Mimeo Database package.

This package exposes databases support required by Mimeo Utils.

It contains the following modules:
* mimeo_db
    The Mimeo Database module.
* cities
    The Cities module.
* countries
    The Countries module.
* currencies
    The Currencies module.
* first_names
    The First Names module.
* last_names
    The Last Names module.
* exc
    The Mimeo Database Exceptions module.

The Mimeo Context package exports the following classes:
* MimeoDB
    Facade class exposing READ operations on all Mimeo CSV data.
* CitiesDB
    Class exposing READ operations on cities CSV data.
* CountriesDB
    Class exposing READ operations on countries CSV data.
* CurrenciesDB
    Class exposing READ operations on currencies CSV data.
* FirstNamesDB
    Class exposing READ operations on forenames CSV data.
* LastNamesDB
    Class exposing READ operations on surnames CSV data.
* City
    DTO class representing a single row in cities CSV data.
* Country
    DTO class representing a single row in countries CSV data.
* Currency
    DTO class representing a single row in currencies CSV data.
* FirstName
    DTO class representing a single row in forenames CSV data.

To use this package, simply import the desired class:
    from mimeo.database import MimeoDB
    from mimeo.database.exc import DataNotFoundError
"""
from __future__ import annotations

from .cities import CitiesDB, City
from .countries import CountriesDB, Country
from .currencies import CurrenciesDB, Currency
from .first_names import FirstName, FirstNamesDB
from .last_names import LastNamesDB
from .mimeo_db import MimeoDB

__all__ = [
    "City",
    "Country",
    "Currency",
    "FirstName",
    "CitiesDB",
    "CountriesDB",
    "CurrenciesDB",
    "FirstNamesDB",
    "LastNamesDB",
    "MimeoDB",
]



"""The Mimeo Utils package.

It contains modules supporting the Mimeo Utils rendering:
* mimeo_utils
    The Mimeo Utils module.
* renderers
    The Mimeo Renderers module.
* exc
    The Mimeo Utils Exceptions module.

This package exports the following classes:
* MimeoUtil
    A superclass for all Mimeo Utils.
* RandomStringUtil
    A MimeoUtil implementation rendering a random string value.
* RandomIntegerUtil
    A MimeoUtil implementation rendering a random integer value.
* RandomItemUtil
    A MimeoUtil implementation rendering a random item.
* DateUtil
    A MimeoUtil implementation rendering a stringified date value.
* DateTimeUtil
    A MimeoUtil implementation rendering a stringified date time value.
* AutoIncrementUtil
    A MimeoUtil implementation rendering an auto incremented ID.
* CurrentIterationUtil
    A MimeoUtil implementation rendering a current iteration ID.
* KeyUtil
    A MimeoUtil implementation rendering a unique identifier.
* CityUtil
    A MimeoUtil implementation rendering city names.
* CountryUtil
    A MimeoUtil implementation rendering country details.
* CurrencyUtil
    A MimeoUtil implementation rendering currency details.
* FirstNameUtil
    A MimeoUtil implementation rendering forenames.
* LastNameUtil
    A MimeoUtil implementation rendering surnames.
* MimeoRenderer
    A Facade class rendering Mimeo Utils, Vars and Special Fields.

To use this package, simply import the desired class:
    from mimeo.utils import MimeoRenderer
"""
from __future__ import annotations

from .mimeo_utils import (AutoIncrementUtil, CityUtil, CountryUtil,
                          CurrencyUtil, CurrentIterationUtil, DateTimeUtil,
                          DateUtil, FirstNameUtil, KeyUtil, LastNameUtil,
                          MimeoUtil, RandomIntegerUtil, RandomItemUtil,
                          RandomStringUtil)
from .renderers import MimeoRenderer

__all__ = ["AutoIncrementUtil", "CityUtil", "CountryUtil", "CurrencyUtil",
           "CurrentIterationUtil", "DateTimeUtil", "DateUtil", "FirstNameUtil",
           "KeyUtil", "LastNameUtil", "MimeoUtil", "RandomIntegerUtil",
           "RandomItemUtil", "RandomStringUtil", "MimeoRenderer"]



"""The Mimeo Tools module.

It is meant to store all useful functions that can help in any
non-Mimeo-specific operations. It exports the following functions:
    * get_resource(resource_name: str) -> TextIO
        Return a Mimeo resource.
"""
from __future__ import annotations

import importlib.resources as pkg_resources
from typing import TextIO

from mimeo import resources as data
from mimeo.resources.exc import ResourceNotFoundError


def get_resource(
        resource_name: str,
) -> TextIO:
    """Return a Mimeo resource.

    The resource needs to be included in mimeo.resources package
    to be returned.

    Parameters
    ----------
    resource_name : str
        A Mimeo resource name

    Returns
    -------
    TextIO
        A Mimeo resource

    Raises
    ------
    ResourceNotFoundError
        If the resource does not exist
    """
    try:
        return pkg_resources.open_text(data, resource_name)
    except FileNotFoundError:
        raise ResourceNotFoundError(resource_name) from FileNotFoundError



"""The Alive module.

It exports classes allowing you to define a class as alive and forbid
to have more than one alive instance at the same time. This
functionality is similar to the Singleton pattern. However, using
OnlyOneAlive class you can instantiate a class multiple times.
Thanks to that a single instance will not store several configurations
depending on execution time. This pattern ensures your instances are
consistent and will be reached whenever needed using a constructor.

The complete pattern is supported using 2 exported classes:
    * OnlyOneAlive
        A type ensuring there's only one instance qualified to be used.
    * Alive
        A superclass for OnlyOneAlive classes.

Example
-------
class SomeClass(Alive, metaclass=OnlyOneAlive):

    def __init__(self, config: dict = None):
        self._config = config

    def get_x(self):
        super().assert_alive()
        return self._config.get('x', None)


with SomeClass({'x': 1}) as alive_1:
    x_1 = alive_1.get_x()

    alive_2 = SomeClass()
    x_2 = alive_2.get_x()

    print('Alive 1:', x_1)  # 1
    print('Alive 2:', x_2)  # 1
"""
from __future__ import annotations

from types import TracebackType
from typing import ClassVar

from mimeo.meta.exc import InstanceNotAliveError


class OnlyOneAlive(type):
    """A type ensuring there's only one instance qualified to be used.

    The OnlyOneAlive type caches all instances being created for each
    Alive subclass. If there is any instance being alive, it is
    returned by a constructor. Otherwise, an Alive subclass is
    instantiated and returned.
    """

    _INSTANCES: ClassVar[dict] = {}

    def __call__(
            cls,
            *args,
            **kwargs,
    ):
        """Ensure there's only one instance qualified to be used."""
        if cls not in cls._INSTANCES:
            cls._INSTANCES[cls] = []

        alive_instance = next((i for i in cls._INSTANCES[cls] if i.is_alive()), None)
        if alive_instance is None:
            instance = super().__call__(*args, **kwargs)
            cls._INSTANCES[cls].append(instance)
            return instance
        return alive_instance


class Alive:
    """A superclass for OnlyOneAlive classes.

    It provides several features defined to use OnlyOneAlive pattern:
    * Context Manager functionality
    * a method checking if an instance is alive
    * a method throwing an exception when the instance is not alive

    Methods
    -------
    is_alive()
        Verify if the instance is alive.
    assert_alive()
        Assert the instance is alive.
    """

    def __init__(
            self,
    ):
        self._alive: bool = False

    def __enter__(
            self,
    ) -> Alive:
        """Enter the Alive instance.

        It sets the internal `alive` attribute to True.

        Returns
        -------
        self : Alive
            A Alive instance
        """
        self._alive = True
        return self

    def __exit__(
            self,
            exc_type: type | None,
            exc_val: BaseException | None,
            exc_tb: TracebackType | None,
    ) -> None:
        """Exit the Alive instance.

        It sets the internal `alive` attribute to False.

        Parameters
        ----------
        exc_type : type | None
            An exception's type
        exc_val : BaseException | None
            An exception's value
        exc_tb  TracebackType | None
            An exception's traceback

        Returns
        -------
        None
            A None value
        """
        self._alive = False

    def assert_alive(
            self,
    ) -> bool:
        """Assert the instance is alive.

        Returns
        -------
        bool
            True

        Raises
        ------
        InstanceNotAliveError
            If the instance is not alive
        """
        if not self.is_alive():
            raise InstanceNotAliveError
        return self.is_alive()

    def is_alive(
            self,
    ) -> bool:
        """Verify if the instance is alive."""
        return self._alive



"""The Mimeo Configuration package.

It contains the following modules:
* mimeo_config
    The Mimeo Configuration module
* constants
    The Mimeo Configuration Constants module
* exc
    The Mimeo Configuration Exceptions module

The Mimeo Configuration package exports a class representing
root Mimeo Configuration component and a factory:
* MimeoConfig
    A MimeoDTO class representing Mimeo Configuration
* MimeoConfigFactory
    A factory class to instantiate a MimeoConfig

To use this package, simply import it:
    from mimeo.config import MimeoConfigFactory, MimeoConfig
    from mimeo.config import constants as cc
    from mimeo.config.exc import UnsupportedPropertyValueError
"""
from __future__ import annotations

from .mimeo_config import MimeoConfig, MimeoConfigFactory

__all__ = ["MimeoConfig", "MimeoConfigFactory"]



"""The First Names module.

It exports classes related to forenames CSV data:
    * FirstName
        DTO class representing a single row in forenames CSV data.
    * FirstNamesDB
        Class exposing READ operations on forenames CSV data.
"""
from __future__ import annotations

from typing import ClassVar

import pandas
from pandas import DataFrame

from mimeo import tools
from mimeo.database.exc import InvalidIndexError, InvalidSexError


class FirstName:
    """DTO class representing a single row in forenames CSV data.

    Attributes
    ----------
    name : str
        A first name
    sex : str
        A sex value
    """

    def __init__(
            self,
            name: str,
            sex: str,
    ):
        """Initialize FirstName class.

        Parameters
        ----------
        name : str
            A first name
        sex : str
            A sex value
        """
        self.name: str = name
        self.sex: str = sex

    def __str__(
            self,
    ) -> str:
        """Stringify the FirstName instance.

        Returns
        -------
        str
            A stringified `dict` representation of the FirstName instance
        """
        return str({
            "name": self.name,
            "sex": self.sex,
        })

    def __repr__(
            self,
    ) -> str:
        """Represent the FirstName instance.

        Returns
        -------
        str
            A python representation of the FirstName instance
        """
        return (f"FirstName("
                f"name='{self.name}', "
                f"sex='{self.sex}')")


class FirstNamesDB:
    """Class exposing READ operations on forenames CSV data.

    Attributes
    ----------
    NUM_OF_RECORDS : int
        A number of rows in forenames CSV data

    Methods
    -------
    get_first_names() -> list[FirstName]
        Get all first names.
    get_first_names_by_sex(sex: str) -> list[FirstName]
        Get first names for a specific sex.
    get_first_name_at(index: int) -> FirstName
        Get a first name at `index` position.
    """

    NUM_OF_RECORDS: int = 7455
    __SUPPORTED_SEX: tuple = ("M", "F")
    __FIRST_NAMES_DB: str = "forenames.csv"
    __FIRST_NAMES_DF: DataFrame = None
    __FIRST_NAMES: ClassVar[list] = None
    __NAMES_FOR_SEX: ClassVar[dict] = {}

    def get_first_name_at(
            self,
            index: int,
    ) -> FirstName:
        """Get a first name at `index` position.

        Parameters
        ----------
        index : int
            A first name row index

        Returns
        -------
        FirstName
            A specific first name

        Raises
        ------
        InvalidIndexError
            If the provided `index` is out of bounds
        """
        first_names = FirstNamesDB._get_first_names()
        try:
            return first_names[index]
        except IndexError:
            last_index = FirstNamesDB.NUM_OF_RECORDS-1
            raise InvalidIndexError(index, last_index) from IndexError

    def get_first_names_by_sex(
            self,
            sex: str,
    ) -> list[FirstName]:
        """Get first names for a specific sex.

        Parameters
        ----------
        sex : str
            A sex value to filter first names

        Returns
        -------
        list[FirstName]
            List of first names filtered by sex

        Raises
        ------
        InvalidSexError
            If the provided `sex` value is not supported
        """
        if sex not in FirstNamesDB.__SUPPORTED_SEX:
            raise InvalidSexError(FirstNamesDB.__SUPPORTED_SEX)
        return FirstNamesDB._get_first_names_by_sex(sex).copy()

    def get_first_names(
            self,
    ) -> list[FirstName]:
        """Get all first names.

        Returns
        -------
        list[FirstName]
            List of all first names
        """
        return FirstNamesDB._get_first_names().copy()

    @classmethod
    def _get_first_names_by_sex(
            cls,
            sex: str,
    ) -> list[FirstName]:
        """Get first names for a specific sex from cache.

        The first names list for sex is initialized for the first time
        and cached in internal class attribute.

        Parameters
        ----------
        sex : str
            A sex value to filter first names

        Returns
        -------
        list[FirstName]
            List of first names filtered by sex
        """
        if sex not in cls.__NAMES_FOR_SEX:
            first_names = cls._get_first_names()
            cls.__NAMES_FOR_SEX[sex] = list(filter(lambda n: n.sex == sex, first_names))
        return cls.__NAMES_FOR_SEX[sex]

    @classmethod
    def _get_first_names(
            cls,
    ) -> list[FirstName]:
        """Get all first names from cache.

        The first names list is initialized for the first time and
        cached in internal class attribute.

        Returns
        -------
        list[FirstName]
            List of all first names
        """
        if cls.__FIRST_NAMES is None:
            cls.__FIRST_NAMES = [FirstName(row.NAME, row.SEX)
                                 for row in cls._get_first_names_df().itertuples()]
        return cls.__FIRST_NAMES

    @classmethod
    def _get_first_names_df(
            cls,
    ) -> pandas.DataFrame:
        """Load forenames CSV data and save in internal class attribute."""
        if cls.__FIRST_NAMES_DF is None:
            data = tools.get_resource(cls.__FIRST_NAMES_DB)
            cls.__FIRST_NAMES_DF = pandas.read_csv(data)
        return cls.__FIRST_NAMES_DF



"""The Mimeo Context module.

It exports only one class:
    * MimeoContext
        A class managing Mimeo-Template-dependent utilities.
"""
from __future__ import annotations

import random

from mimeo.context import MimeoIteration
from mimeo.context.exc import (ContextIterationNotFoundError,
                               MinimumIdentifierReachedError,
                               UninitializedContextIterationError)
from mimeo.database import MimeoDB
from mimeo.database.exc import DataNotFoundError, OutOfStockError


class MimeoContext:
    """A class managing Mimeo-Template-dependent utilities.

    It allows you to reach a specific iteration of a template
    generation, and ensures uniqueness of all values generated
    by Mimeo Utils (and supporting this feature). Giving access
    to all iterations, it allows you to get a key or any special field
    of them.

    Attributes
    ----------
    name : str
        A context name (a model's root name if not explicitly
        defined in a Mimeo Configuration)

    Methods
    -------
    next_id() -> int
        Increment an identifier and return the current (incremented) one.
    curr_id() -> int
        Return the current identifier within the context.
    prev_id() -> int
        Decrement an identifier and return the current (decremented) one.
    next_iteration() -> MimeoIteration
        Initialize a next iteration within the context.
    curr_iteration() -> MimeoIteration
        Return the current iteration within the context.
    get_iteration(iteration_id: int) -> MimeoIteration
        Return a specific iteration from the context.
    clear_iterations()
        Clear out all context iterations.
    next_country_index() -> int
        Provide next unique country index.
    next_city_index(country: str = None) -> int
        Provide next unique city index.
    next_currency_index() -> int
        Provide next unique currency index.
    next_first_name_index(sex: str = None) -> int
        Provide next unique first name index.
    next_last_name_index() -> int
        Provide next unique last name index.
    """

    _ALL: str = "_ALL_"
    _INITIAL_COUNT: str = "init-count"
    _INDEXES: str = "indexes"

    def __init__(
            self,
            name: str,
    ):
        """Initialize MimeoContext class.

        Parameters
        ----------
        name : str
            A context name
        """
        self.name: str = name
        self._id: int = 0
        self._iterations: list[MimeoIteration] = []
        self._countries_indexes: list[int] | None = None
        self._cities_indexes: dict = {}
        self._currencies_indexes: list[int] | None = None
        self._first_names_indexes: dict = {}
        self._last_names_indexes: list[int] | None = None

    def next_id(
            self,
    ) -> int:
        """Increment an identifier and return the current (incremented) one.

        Identifier is used by Auto Increment Mimeo Util.

        Returns
        -------
        int
            A next identifier within the context
        """
        self._id += 1
        return self.curr_id()

    def curr_id(
            self,
    ) -> int:
        """Return the current identifier within the context.

        Identifier is used by Auto Increment Mimeo Util.

        Returns
        -------
        int
            The current identifier within the context
        """
        return self._id

    def prev_id(
            self,
    ) -> int:
        """Decrement an identifier and return the current (decremented) one.

        Identifier is used by Auto Increment Mimeo Util.
        This method is meant to be used when an error appear
        after incrementation.

        Returns
        -------
        int
            A previous identifier within the context

        Raises
        ------
        MinimumIdentifierReachedError
            If the current identifier (before decrement) equals 0
        """
        if self._id == 0:
            raise MinimumIdentifierReachedError
        self._id -= 1
        return self.curr_id()

    def next_iteration(
            self,
    ) -> MimeoIteration:
        """Initialize a next iteration within the context.

        To initialize the iteration, it gets the last iteration id and
        provides its incrementation.

        Returns
        -------
        next_iteration : MimeoIteration
            The initialized iteration
        """
        if len(self._iterations) == 0:
            next_iteration_id = 1
        else:
            next_iteration_id = self._iterations[-1].id + 1
        next_iteration = MimeoIteration(next_iteration_id)
        self._iterations.append(next_iteration)
        return next_iteration

    def curr_iteration(
            self,
    ) -> MimeoIteration:
        """Return the current iteration within the context.

        Returns
        -------
        MimeoIteration
            The current iteration within the context

        Raises
        ------
        UninitializedContextIterationError
            If no iteration has been initialized yet for the context
        """
        if len(self._iterations) == 0:
            raise UninitializedContextIterationError(self.name)
        return self._iterations[-1]

    def get_iteration(
            self,
            iteration_id: int,
    ) -> MimeoIteration:
        """Return a specific iteration from the context.

        Returns
        -------
        int
            A specific iteration

        Raises
        ------
        ContextIterationNotFoundError
            If the context does not have an iteration with the id
            provided
        """
        iteration = next(filter(lambda i: i.id == iteration_id, self._iterations), None)
        if iteration is None:
            raise ContextIterationNotFoundError(iteration_id, self.name)
        return iteration

    def clear_iterations(
            self,
    ):
        """Clear out all context iterations.

        This method is meant to be used in case of nested templates.
        Thanks to iteration reset the nested template is properly
        generated in context of the next parent template's iteration.
        """
        self._iterations = []

    def next_country_index(
            self,
    ) -> int:
        """Provide next unique country index.

        When used for the first time in the specific context
        it populates internal countries' indexes list. This approach
        ensures country uniqueness without time-consuming operations.
        Each time it verifies if the internal list still contains some
        indexes.
        This method is used by the Country Mimeo Util to get a country
        entry at a specific index in database.

        Returns
        -------
        int
            Next unique country identifier

        Raises
        ------
        OutOfStockError
            If all countries' indexes have been consumed already
        """
        self._initialize_countries_indexes()
        self._validate_countries()

        return self._countries_indexes.pop()

    def next_city_index(
            self,
            country: str | None = None,
    ) -> int:
        """Provide next unique city index.

        When used for the first time in the specific context
        it populates internal cities' indexes map. Each `country` key
        has its own list initialized as same as country-agnostic one.
        This approach ensures city uniqueness without time-consuming
        operations. Each time it verifies if the internal list still
        contains some indexes.
        This method is used by the City Mimeo Util to get a city entry
        at a specific index in database.

        Parameters
        ----------
        country : str, default _ALL_
            A country limitation to find cities. When None - all
            countries are considered.

        Returns
        -------
        int
            Next unique city identifier

        Raises
        ------
        DataNotFoundError
            If database does not contain any cities for the provided
            `country`
        OutOfStockError
            If all cities' indexes have been consumed already
        """
        country = country if country is not None else MimeoContext._ALL
        self._initialize_cities_indexes(country)
        self._validate_cities(country)

        return self._cities_indexes[country][MimeoContext._INDEXES].pop()

    def next_currency_index(
            self,
    ) -> int:
        """Provide next unique currency index.

        When used for the first time in the specific context
        it populates internal currencies' indexes list. This approach
        ensures country uniqueness without time-consuming operations.
        Each time it verifies if the internal list still contains some
        indexes.
        This method is used by the Currency Mimeo Util to get a currency
        entry at a specific index in database.

        Returns
        -------
        int
            Next unique currency identifier

        Raises
        ------
        OutOfStockError
            If all currencies' indexes have been consumed already
        """
        self._initialize_currencies_indexes()
        self._validate_currencies()

        return self._currencies_indexes.pop()

    def next_first_name_index(
            self,
            sex: str | None = None,
    ) -> int:
        """Provide next unique first name index.

        When used for the first time in the specific context
        it populates internal first names' indexes map. Each `sex` key
        has its own list initialized as same as sex-agnostic one.
        This approach ensures forename uniqueness without
        time-consuming operations. Each time it verifies if the internal
        list still contains some indexes.
        This method is used by the First Name Mimeo Util to get a
        first name entry at a specific index in database.

        Parameters
        ----------
        sex : str, default _ALL_
            A sex limitation to find names. When None - both
            sexes are considered.

        Returns
        -------
        int
            Next unique first name identifier

        Raises
        ------
        InvalidSexError
            If `sex` is not 'M' nor 'F' value
        OutOfStockError
            If all first names' indexes have been consumed already
        """
        sex = sex if sex is not None else MimeoContext._ALL
        self._initialize_first_names_indexes(sex)
        self._validate_first_names(sex)

        return self._first_names_indexes[sex][MimeoContext._INDEXES].pop()

    def next_last_name_index(
            self,
    ) -> int:
        """Provide next unique last name index.

        When used for the first time in the specific context
        it populates internal last names' indexes list. This approach
        ensures surnames uniqueness without time-consuming operations.
        Each time it verifies if the internal list still contains some
        indexes.
        This method is used by the Last Name Mimeo Util to get a
        last name entry at a specific index in database.

        Returns
        -------
        int
            Next unique last name identifier

        Raises
        ------
        OutOfStockError
            If all last names' indexes have been consumed already
        """
        self._initialize_last_names_indexes()
        self._validate_last_names()

        return self._last_names_indexes.pop()

    def _initialize_countries_indexes(
            self,
    ):
        """Initialize countries' indexes with unique integers.

        The list length and range depends on the number of country
        records in database.
        """
        if self._countries_indexes is None:
            num_of_entries = MimeoDB.NUM_OF_COUNTRIES
            countries_indexes = random.sample(range(num_of_entries), num_of_entries)
            self._countries_indexes = countries_indexes

    def _initialize_cities_indexes(
            self,
            country: str,
    ):
        """Initialize cities' indexes with unique integers.

        The list length and range depends on the number of city
        records in database for the `country`.

        Raises
        ------
        DataNotFoundError
            If database does not contain any cities for the provided
            `country`
        """
        if country not in self._cities_indexes:
            if country == MimeoContext._ALL:
                num_of_entries = MimeoDB.NUM_OF_CITIES
            else:
                country_cities = MimeoDB().get_cities_of(country)
                num_of_entries = len(country_cities)
                if num_of_entries == 0:
                    raise DataNotFoundError(DataNotFoundError.Code.ERR_2,
                                            data="city",
                                            param_name="country",
                                            param_val=country)

            cities_indexes = random.sample(range(num_of_entries), num_of_entries)
            self._cities_indexes[country] = {
                MimeoContext._INITIAL_COUNT: num_of_entries,
                MimeoContext._INDEXES: cities_indexes,
            }

    def _initialize_currencies_indexes(
            self,
    ):
        """Initialize currencies' indexes with unique integers.

        The list length and range depends on the number of currency
        records in database.
        """
        if self._currencies_indexes is None:
            num_of_entries = MimeoDB.NUM_OF_CURRENCIES
            currencies_indexes = random.sample(range(num_of_entries), num_of_entries)
            self._currencies_indexes = currencies_indexes

    def _initialize_first_names_indexes(
            self,
            sex: str,
    ):
        """Initialize first names' indexes with integers.

        The list length and range depends on the number of first name
        records in database for the `sex`.

        Raises
        ------
        InvalidSexError
            If `sex` is not 'M' nor 'F' value
        """
        if sex not in self._first_names_indexes:
            if sex == MimeoContext._ALL:
                num_of_entries = MimeoDB.NUM_OF_FIRST_NAMES
            else:
                first_names_for_sex = MimeoDB().get_first_names_by_sex(sex)
                num_of_entries = len(first_names_for_sex)

            first_names_indexes = random.sample(range(num_of_entries), num_of_entries)
            self._first_names_indexes[sex] = {
                MimeoContext._INITIAL_COUNT: num_of_entries,
                MimeoContext._INDEXES: first_names_indexes,
            }

    def _initialize_last_names_indexes(
            self,
    ):
        """Initialize last names' indexes with unique integers.

        The list length and range depends on the number of last name
        records in database.
        """
        if self._last_names_indexes is None:
            num_of_entries = MimeoDB.NUM_OF_LAST_NAMES
            last_names_indexes = random.sample(range(num_of_entries), num_of_entries)
            self._last_names_indexes = last_names_indexes

    def _validate_countries(
            self,
    ):
        """Verify if all countries' indexes have been consumed.

        Raises
        ------
        OutOfStockError
            If all countries' indexes have been consumed already
        """
        if len(self._countries_indexes) == 0:
            raise OutOfStockError(OutOfStockError.Code.ERR_1,
                                  num=MimeoDB.NUM_OF_COUNTRIES,
                                  data="countries")

    def _validate_cities(
            self,
            country: str,
    ):
        """Verify if all cities' indexes have been consumed.

        Raises
        ------
        OutOfStockError
            If all cities' indexes have been consumed already
        """
        if len(self._cities_indexes[country][MimeoContext._INDEXES]) == 0:
            init_count = self._cities_indexes[country][MimeoContext._INITIAL_COUNT]
            if country == MimeoContext._ALL:
                raise OutOfStockError(OutOfStockError.Code.ERR_1,
                                      num=init_count,
                                      data="cities")
            raise OutOfStockError(OutOfStockError.Code.ERR_2,
                                  num=init_count,
                                  data="cities",
                                  param_val=country)

    def _validate_currencies(
            self,
    ):
        """Verify if all currencies' indexes have been consumed.

        Raises
        ------
        OutOfStockError
            If all currencies' indexes have been consumed already
        """
        if len(self._currencies_indexes) == 0:
            raise OutOfStockError(OutOfStockError.Code.ERR_1,
                                  num=MimeoDB.NUM_OF_CURRENCIES,
                                  data="currencies")

    def _validate_first_names(
            self,
            sex: str,
    ):
        """Verify if all first names' indexes have been consumed.

        Raises
        ------
        OutOfStockError
            If all first names' indexes have been consumed already
        """
        if len(self._first_names_indexes[sex][MimeoContext._INDEXES]) == 0:
            init_count = self._first_names_indexes[sex][MimeoContext._INITIAL_COUNT]
            if sex == MimeoContext._ALL:
                sex_info = ""
            elif sex == "M":
                sex_info = "male "
            else:
                sex_info = "female "
            raise OutOfStockError(OutOfStockError.Code.ERR_1,
                                  num=init_count,
                                  data=f"{sex_info}first names")

    def _validate_last_names(
            self,
    ):
        """Verify if all last names' indexes have been consumed.

        Raises
        ------
        OutOfStockError
            If all last names' indexes have been consumed already
        """
        if len(self._last_names_indexes) == 0:
            raise OutOfStockError(OutOfStockError.Code.ERR_1,
                                  num=MimeoDB.NUM_OF_LAST_NAMES,
                                  data="last names")



"""The Mimeo Context Decorators module.

It defines the following decorators:

- @mimeo_context:
    Decorator providing a 'context' parameter to a function
- @mimeo_context_switch:
    Decorator switching context before and after function's call
- @mimeo_next_iteration:
    Decorator incrementing current context's iteration
- @mimeo_clear_iterations:
    Decorator clearing iterations of the current context

"""
from __future__ import annotations

import functools
from typing import Callable

from mimeo.config.mimeo_config import MimeoTemplate
from mimeo.context import MimeoContext, MimeoContextManager


def mimeo_context(
        func: Callable,
) -> Callable:
    """Provide a 'context' parameter to a function.

    It can be used only for functions having defined a 'context'
    parameter. In case the function being decorated is called providing
    this param, decorator simply calls it without any modification.
    Otherwise - and this is the main purpose of the @mimeo_context
    decorator - it provides the current Mimeo Context.

    Parameters
    ----------
    func : Callable
        The function being decorated

    Returns
    -------
    inject_context : Callable
        The decorated function
    """

    @functools.wraps(func)
    def inject_context(
            *args,
            **kwargs,
    ):
        if any(isinstance(arg, MimeoContext) for arg in args) or "context" in kwargs:
            result = func(*args, **kwargs)
        else:
            current_ctx = MimeoContextManager().get_current_context()
            result = func(*args, **kwargs, context=current_ctx)
        return result

    return inject_context


def mimeo_context_switch(
        func: Callable,
) -> Callable:
    """Switch context before and after function's call.

    It can be used only for functions having defined a MimeoTemplate
    parameter. It switches a Mimeo Context to the one defined in
    Mimeo Template, and after the function's being decorated call
    switches back to the previous one. It helps to handle nested
    Mimeo Templates. This decorator meant to be used for
    Generator's function that generates data from a template.

    Parameters
    ----------
    func : Callable
        The function being decorated

    Returns
    -------
    switch_context : Callable
        The decorated function
    """

    @functools.wraps(func)
    def switch_context(
            *args,
            **kwargs,
    ):
        context_mng = MimeoContextManager()
        prev_context = context_mng.get_current_context()

        if "template" in kwargs:
            template = kwargs["template"]
        else:
            template = next(arg for arg in args if isinstance(arg, MimeoTemplate))
        context_name = template.model.context_name
        next_context = context_mng.get_context(context_name)
        context_mng.set_current_context(next_context)
        result = func(*args, **kwargs)

        context_mng.set_current_context(prev_context)
        return result

    return switch_context


def mimeo_next_iteration(
        func: Callable,
):
    """Increment current context's iteration.

    It is meant to be used for Generator's function that
    generates a single data copy.

    Parameters
    ----------
    func : Callable
        The function being decorated

    Returns
    -------
    next_iteration : Callable
        The decorated function
    """

    @functools.wraps(func)
    def next_iteration(
            *args,
            **kwargs,
    ):
        MimeoContextManager().get_current_context().next_iteration()
        return func(*args, **kwargs)

    return next_iteration


def mimeo_clear_iterations(
        func: Callable,
):
    """Clear iterations of the current context.

    It is meant to be used for Generator's function that generates
    data from a template.

    Parameters
    ----------
    func : Callable
        The function being decorated

    Returns
    -------
    clear_iterations : Callable
        The decorated function
    """

    @functools.wraps(func)
    def clear_iterations(
            *args,
            **kwargs,
    ):
        MimeoContextManager().get_current_context().clear_iterations()
        return func(*args, **kwargs)

    return clear_iterations



"""The Mimeo Utils Exceptions module.

It contains all custom exceptions related to Mimeo Utils:
    * InvalidMimeoUtilError
        A custom Exception class for an invalid Mimeo Util.
    * InvalidMimeoUtilError.Code
        An Enumeration class for InvalidMimeoUtilError error codes.
    * InvalidValueError
        A custom Exception class for an invalid value in Mimeo Util.
    * InvalidValueError.Code
        An Enumeration class for InvalidValueError error codes.
    * NotASpecialFieldError
        A custom Exception class for a field used as a special.
"""


from __future__ import annotations

from enum import Enum


class InvalidMimeoUtilError(Exception):
    """A custom Exception class for an invalid Mimeo Util.

    Raised when Mimeo Util node has missing _name property, or it
    does not match any Mimeo Util.
    """

    class Code(Enum):
        """An Enumeration class for InvalidMimeoUtilError error codes.

        Attributes
        ----------
        ERR_1: str
            An error code for a missing Mimeo Util name in configuration
        ERR_2: str
            An error code for an unsupported Mimeo Util
        """

        ERR_1: str = "MISSING_NAME"
        ERR_2: str = "UNSUPPORTED_MIMEO_UTIL"

    def __init__(
            self,
            code: InvalidMimeoUtilError.Code,
            **kwargs,
    ):
        """Initialize InvalidMimeoUtilError exception with details.

        Extends Exception constructor with a custom message. The message depends on
        an internal InvalidMimeoUtilError code.

        Parameters
        ----------
        code : InvalidMimeoUtilError.Code
            An internal error code
        kwargs
            An error details
        """
        msg = self._get_msg(code, kwargs)
        super().__init__(msg)

    @classmethod
    def _get_msg(
            cls,
            code: InvalidMimeoUtilError.Code,
            details: dict,
    ):
        """Return a custom message based on an error code.

        Parameters
        ----------
        code : InvalidMimeoUtilError.Code
            An internal error code
        details : dict
            An error details

        Returns
        -------
        str
            A custom error message

        Raises
        ------
        ValueError
            If the code argument is not InvalidMimeoUtilError.Code enum
        """
        if code == cls.Code.ERR_1:
            return f"Missing Mimeo Util name in configuration [{details['config']}]!"
        if code == cls.Code.ERR_2:
            return f"No such Mimeo Util [{details['name']}]!"

        msg = f"Provided error code is not a {cls.__name__}.Code enum!"
        raise ValueError(msg)


class InvalidValueError(Exception):
    """A custom Exception class for an invalid value in Mimeo Util.

    Raised when Mimeo Util node is incorrectly parametrized.
    """

    class Code(Enum):
        """An Enumeration class for InvalidValueError error codes.

        Attributes
        ----------
        ERR_1: str
            An error code for a negative length in Mimeo Util configuration
        ERR_2: str
            An error code for a limit param lower than start
        ERR_3: str
            An error code for an invalid param type
        ERR_3: str
            An error code for an unsupported value
        """

        ERR_1: str = "NEGATIVE_LENGTH"
        ERR_2: str = "LIMIT_LOWER_THAN_START"
        ERR_3: str = "INVALID_TYPE"
        ERR_4: str = "UNSUPPORTED_VALUE"

    def __init__(
            self,
            code: InvalidValueError.Code,
            **kwargs,
    ):
        """Initialize InvalidValueError exception with details.

        Extends Exception constructor with a custom message. The message depends on
        an internal InvalidValueError code.

        Parameters
        ----------
        code : InvalidValueError.Code
            An internal error code
        kwargs
            An error details
        """
        msg = self._get_msg(code, kwargs)
        super().__init__(msg)

    @classmethod
    def _get_msg(
            cls,
            code: InvalidMimeoUtilError.Code,
            details: dict,
    ):
        """Return a custom message based on an error code.

        Parameters
        ----------
        code : InvalidValueError.Code
            An internal error code
        details : dict
            An error details

        Returns
        -------
        str
            A custom error message

        Raises
        ------
        ValueError
            If the code argument is not InvalidValueError.Code enum
        """
        if code == cls.Code.ERR_1:
            msg = (f"The {details['util']} Mimeo Util cannot be parametrized with "
                   f"negative length [{details['length']}] value")
        elif code == cls.Code.ERR_2:
            msg = (f"The {details['util']} Mimeo Util cannot be parametrized with "
                   f"limit [{details['limit']}] lower than start [{details['start']}]")
        elif code == cls.Code.ERR_3:
            msg = (f"The {details['util']} Mimeo Util require a {details['type']} "
                   f"value for the {details['param_name']} parameter and was: "
                   f"[{details['param_val']}].")
        elif code == cls.Code.ERR_4:
            msg = (f"The {details['util']} Mimeo Util does not support a value "
                   f"[{details['value']}]. "
                   f"Supported values are: {', '.join(details['supported_values'])}.")
        else:
            msg = f"Provided error code is not a {cls.__name__}.Code enum!"
            raise ValueError(msg)
        return msg


class NotASpecialFieldError(Exception):
    """A custom Exception class for a field used as a special.

    Raised while attempting to retrieve special field name when it is
    not a special one.
    """

    def __init__(
            self,
            field_name: str,
    ):
        """Initialize NotASpecialFieldError exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        field_name : str
            A field name
        """
        msg = f"Provided field [{field_name}] is not a special one (use {':NAME:'})!"
        super().__init__(msg)



"""The Mimeo Database Exceptions module.

It contains all custom exceptions related to Mimeo Context:
    * InvalidIndexError
        A custom Exception class for invalid row index.
    * InvalidSexError
        A custom Exception class for uninitialized context's iteration.
    * DataNotFoundError
        A custom Exception class for not found data.
    * DataNotFoundError.Code
        An Enumeration class for DataNotFoundError error codes.
    * OutOfStockError
        A custom Exception class for invalid special field's name.
    * DataNotFoundError.Code
        An Enumeration class for OutOfStockError error codes.
"""


from __future__ import annotations

from enum import Enum


class InvalidIndexError(Exception):
    """A custom Exception class for invalid row index.

    Raised while attempting to get data at row that does not exist.
    """

    def __init__(
            self,
            index: int,
            last_index: int,
    ):
        """Initialize InvalidIndexError exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        index : int
            An invalid index
        last_index : str
            The last existing index
        """
        msg = f"Provided index [{index}] is out or the range: 0-{last_index}!"
        super().__init__(msg)


class InvalidSexError(Exception):
    """A custom Exception class for invalid sex.

    Raised when sex provided to filter forenames is not supported.
    """

    def __init__(
            self,
            supported_sex_list: tuple,
    ):
        """Initialize InvalidSexError exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        supported_sex_list : int
            A list of supported sex values
        """
        super().__init__(f"Invalid sex (use {' / '.join(supported_sex_list)})!")


class DataNotFoundError(Exception):
    """A custom Exception class for not found data.

    Raised while attempting to filter data by a column's value that
    does not match any row.
    """

    class Code(Enum):
        """An Enumeration class for DataNotFoundError error codes.

        Attributes
        ----------
        ERR_1: str
            An error code for not found data
        ERR_2: str
            An error code for not found data depending on a custom param
        """

        ERR_1 = "NOT_FOUND"
        ERR_2 = "NOT_FOUND_FOR"

    def __init__(
            self,
            code: DataNotFoundError.Code,
            **kwargs,
    ):
        """Initialize DataNotFoundError exception with details.

        Extends Exception constructor with a custom message. The message depends on
        an internal DataNotFoundError code.

        Parameters
        ----------
        code : DataNotFoundError.Code
            An internal error code
        kwargs
            An error details
        """
        msg = self._get_msg(code, kwargs)
        super().__init__(msg)

    @classmethod
    def _get_msg(
            cls,
            code: DataNotFoundError.Code,
            details: dict,
    ):
        """Return a custom message based on an error code.

        Parameters
        ----------
        code : DataNotFoundError.Code
            An internal error code
        details : dict
            An error details

        Returns
        -------
        str
            A custom error message

        Raises
        ------
        ValueError
            If the code argument is not DataNotFoundError.Code enum
        """
        if code == cls.Code.ERR_1:
            return (f"Mimeo database doesn't contain a {details['data']} "
                    f"[{details['value']}].")
        if code == cls.Code.ERR_2:
            return (f"Mimeo database doesn't contain any {details['data']} of "
                    f"the provided {details['param_name']} [{details['param_val']}].")

        msg = f"Provided error code is not a {cls.__name__}.Code enum!"
        raise ValueError(msg)


class OutOfStockError(Exception):
    """A custom Exception class for consuming all data.

    Raised while attempting to get next unique value when all were
    consumed already.
    """

    class Code(Enum):
        """An Enumeration class for OutOfStockError error codes.

        Attributes
        ----------
        ERR_1: str
            An error code for no more unique values
        ERR_2: str
            An error code for no more unique values depending on a custom param
        """

        ERR_1 = "NO_MORE_UNIQUE_VALUES"
        ERR_2 = "NO_MORE_UNIQUE_VALUES_FOR"

    def __init__(
            self,
            code: OutOfStockError.Code,
            **kwargs,
    ):
        """Initialize OutOfStockError exception with details.

        Extends Exception constructor with a custom message. The message depends on
        an internal OutOfStockError code.

        Parameters
        ----------
        code : OutOfStockError.Code
            An internal error code
        kwargs
            An error details
        """
        msg = self._get_msg(code, kwargs)
        super().__init__(msg)

    @classmethod
    def _get_msg(
            cls,
            code: OutOfStockError.Code,
            details: dict,
    ):
        """Return a custom message based on an error code.

        Parameters
        ----------
        code : OutOfStockError.Code
            An internal error code
        details : dict
            An error details

        Returns
        -------
        str
            A custom error message

        Raises
        ------
        ValueError
            If the code argument is not OutOfStockError.Code enum
        """
        if code == cls.Code.ERR_1:
            return (f"No more unique values, "
                    f"database contain only {details['num']} {details['data']}.")
        if code == cls.Code.ERR_2:
            return (f"No more unique values, "
                    f"database contain only {details['num']} {details['data']} of "
                    f"{details['param_val']}.")

        msg = f"Provided error code is not a {cls.__name__}.Code enum!"
        raise ValueError(msg)



"""The Mimeo Context package.

This package provides support fot Mimeo-Template-dependent utilities.

It contains the following modules:
* mimeo_context_manager
    The Mimeo Context Manager module.
* mimeo_context
    The Mimeo Context module.
* mimeo_iteration
    The Mimeo Iteration module.
* decorators
    The Mimeo Context Decorators module.
* exc
    The Mimeo Context Exceptions module.

The Mimeo Context package exports the following classes:
- MimeoIteration:
    A class representing a single iteration in a Mimeo Template.
- MimeoContext:
    A class managing Mimeo-Template-dependent utilities.
- MimeoContextManager:
    An OnlyOneAlive class managing Mimeo Contexts.

To use this package, simply import the desired class:
    from mimeo.context import MimeoContextManager
"""
from .mimeo_iteration import MimeoIteration
from .mimeo_context import MimeoContext
from .mimeo_context_manager import MimeoContextManager

__all__ = ["MimeoIteration", "MimeoContext", "MimeoContextManager"]



"""The Mimeo Renderers module.

This module contains classes useful in value rendering. It exports
the following renderers:
    * MimeoRenderer
        A Facade class rendering Mimeo Utils, Vars and Special Fields.
    * UtilsRenderer
        A class rendering Mimeo Utils.
    * VarsRenderer
        A class rendering Mimeo Vars.
    * RefsRenderer
        A class rendering Mimeo Refs.
    * SpecialFieldsRenderer
        A class rendering Mimeo Special Fields.
"""
from __future__ import annotations

import logging
import re
from typing import Any, ClassVar, Pattern

from mimeo.config import constants as cc
from mimeo.context import MimeoContext, MimeoContextManager
from mimeo.context.decorators import mimeo_context
from mimeo.utils import (AutoIncrementUtil, CityUtil, CountryUtil,
                         CurrencyUtil, CurrentIterationUtil, DateTimeUtil,
                         DateUtil, FirstNameUtil, KeyUtil, LastNameUtil,
                         MimeoUtil, RandomIntegerUtil, RandomItemUtil,
                         RandomStringUtil)
from mimeo.utils.exc import InvalidMimeoUtilError, NotASpecialFieldError

logger = logging.getLogger(__name__)


class UtilsRenderer:
    """A class rendering Mimeo Utils.

    It contains only class methods.

    Methods
    -------
    render_raw(mimeo_util_key: str) -> Any
        Render a Mimeo Util in a raw form.
    render_parametrized(mimeo_util_config: dict) -> Any
        Render a Mimeo Util in a parametrized form.
    """

    MIMEO_UTILS: ClassVar[dict] = {
        RandomStringUtil.KEY: RandomStringUtil,
        RandomIntegerUtil.KEY: RandomIntegerUtil,
        RandomItemUtil.KEY: RandomItemUtil,
        DateUtil.KEY: DateUtil,
        DateTimeUtil.KEY: DateTimeUtil,
        AutoIncrementUtil.KEY: AutoIncrementUtil,
        CurrentIterationUtil.KEY: CurrentIterationUtil,
        KeyUtil.KEY: KeyUtil,
        CityUtil.KEY: CityUtil,
        CountryUtil.KEY: CountryUtil,
        CurrencyUtil.KEY: CurrencyUtil,
        FirstNameUtil.KEY: FirstNameUtil,
        LastNameUtil.KEY: LastNameUtil,
    }
    _INSTANCES: ClassVar[dict] = {}

    @classmethod
    def render_raw(
            cls,
            mimeo_util_key: str,
    ) -> Any:
        """Render a Mimeo Util in a raw form.

        Parameters
        ----------
        mimeo_util_key : str
            A Mimeo Util key (name)

        Returns
        -------
        Any
            Rendered Mimeo Util value.

        Raises
        ------
        InvalidMimeoUtilError
            If the Mimeo Util name does not match any existing Mimeo
            Util.
        """
        return cls.render_parametrized({cc.MODEL_MIMEO_UTIL_NAME_KEY: mimeo_util_key})

    @classmethod
    def render_parametrized(
            cls,
            mimeo_util_config: dict,
    ) -> Any:
        """Render a Mimeo Util in a parametrized form.

        Parameters
        ----------
        mimeo_util_config : dict
            A Mimeo Util configuration

        Returns
        -------
        Any
            Rendered Mimeo Util value.

        Raises
        ------
        InvalidMimeoUtilError
            If the Mimeo Util configuration does not include Mimeo
            Util name, or the parametrized name does not match any
            existing Mimeo Util.
        InvalidValueError
            If a Mimeo Util is incorrectly parametrized.
        InvalidSexError
            If the First Name Mimeo Util has not supported `sex`
            parameter value assigned.
        """
        logger.fine("Rendering a mimeo util [%s]", mimeo_util_config)
        mimeo_util = cls._get_mimeo_util(mimeo_util_config)
        return mimeo_util.render()

    @classmethod
    def _get_mimeo_util(
            cls,
            config: dict,
    ) -> MimeoUtil:
        """Get a Mimeo Util instance based on the configuration.

        All instances are cached to not re-create a Util with the same
        parameters. This method instantiate a Mimeo Util for the first
        time and use the one in the future.

        Parameters
        ----------
        config : dict
            A Mimeo Util configuration

        Returns
        -------
        MimeoUtil
            A Mimeo Util instance

        Raises
        ------
        InvalidMimeoUtilError
            If the Mimeo Util configuration does not include Mimeo
            Util name, or the parametrized name does not match any
            existing Mimeo Util.
        """
        cache_key = cls._generate_cache_key(config)
        if cache_key not in cls._INSTANCES:
            return cls._instantiate_mimeo_util(cache_key, config)
        return cls._INSTANCES.get(cache_key)

    @staticmethod
    def _generate_cache_key(
            config: dict,
    ) -> str:
        """Generate an internal Mimeo Util key from its parameters.

        This method ensures that Mimeo Util instances are cached
        properly for the same parameters.

        Parameters
        ----------
        config : dict
            A Mimeo Util configuration

        Returns
        -------
        str
            An internal Mimeo Util cache key
        """
        return "-".join(":".join([key, str(val)]) for key, val in config.items())

    @classmethod
    def _instantiate_mimeo_util(
            cls,
            cache_key: str,
            config: dict,
    ) -> MimeoUtil:
        """Instantiate a Mimeo Util based on the configuration.

        After instantiation the Mimeo Util is cached for the future.

        Parameters
        ----------
        cache_key : str
            An internal Mimeo Util cache key
        config : dict
            A Mimeo Util configuration

        Returns
        -------
        MimeoUtil
            A Mimeo Util instance

        Raises
        ------
        InvalidMimeoUtilError
            If the Mimeo Util configuration does not include Mimeo
            Util name, or the parametrized name does not match any
            existing Mimeo Util.
        """
        mimeo_util_name = cls.get_mimeo_util_name(config)
        mimeo_util = cls.MIMEO_UTILS.get(mimeo_util_name)(**config)
        cls._INSTANCES[cache_key] = mimeo_util
        return mimeo_util

    @classmethod
    def get_mimeo_util_name(
            cls,
            config: dict,
    ) -> str:
        """Return a verified Mimeo Util name.

        Parameters
        ----------
        config : dict
            A Mimeo Util configuration

        Returns
        -------
        str
            A Mimeo Util name

        Raises
        ------
        InvalidMimeoUtilError
            If the Mimeo Util configuration does not include Mimeo
            Util name, or the parametrized name does not match any
            existing Mimeo Util.
        """
        mimeo_util_name = config.get(cc.MODEL_MIMEO_UTIL_NAME_KEY)
        if mimeo_util_name is None:
            code = InvalidMimeoUtilError.Code.ERR_1
            raise InvalidMimeoUtilError(code, config=config)
        if mimeo_util_name not in cls.MIMEO_UTILS:
            code = InvalidMimeoUtilError.Code.ERR_2
            raise InvalidMimeoUtilError(code, name=mimeo_util_name)
        return mimeo_util_name


class VarsRenderer:
    """A class rendering Mimeo Vars.

    It contains only a class method.

    Methods
    -------
    render(var: str) -> Any
        Render a Mimeo Var.
    """

    @classmethod
    def render(
            cls,
            var: str,
    ) -> str | int | bool | dict:
        """Render a Mimeo Var.

        Parameters
        ----------
        var : str
            A variable name

        Returns
        -------
        Any
            A variable value

        Raises
        ------
        InstanceNotAliveError
            If the MimeoContextManager instance is not alive
        VarNotFoundError
            If the Mimeo Var with the `var` provided does not exist
        """
        logger.fine("Rendering a variable [%s]", var)
        return MimeoContextManager().get_var(var)


class RefsRenderer:
    """A class rendering Mimeo Refs.

    It contains only a class method.

    Methods
    -------
    render(var: str) -> Any
        Render a Mimeo Ref.
    """

    @classmethod
    def render(
            cls,
            ref_name: str,
    ) -> str | int | bool | dict:
        """Render a Mimeo Ref.

        Parameters
        ----------
        ref_name : str
            A reference name

        Returns
        -------
        str | int | bool | dict
            A reference value

        Raises
        ------
        InstanceNotAliveError
            If the MimeoContextManager instance is not alive
        ReferenceNotFoundError
            If there's such a reference configured
        NonPopulatedReferenceError
            If the reference has no values
        NoCorrespondingReferenceError
            If there was no value cached in a corresponding iteration of the source
            context
        """
        logger.fine("Rendering a reference [%s]", ref_name)
        return MimeoContextManager().get_ref(ref_name)


class SpecialFieldsRenderer:
    """A class rendering Mimeo Special Fields.

    It contains only a class method.

    Methods
    -------
    render(field_name: str, context: MimeoContext = None) -> Any
        Render a Mimeo Special Field.
    """

    @classmethod
    @mimeo_context
    def render(
            cls,
            field: str,
            context: MimeoContext | None = None,
    ) -> str | int | bool:
        """Render a Mimeo Special Field.

        Parameters
        ----------
        field : str
            A special field name
        context : MimeoContext, default None
            A current Mimeo Context injected by a decorator

        Returns
        -------
        str | int | bool

        Raises
        ------
        UninitializedContextIterationError
            If no iteration has been initialized yet for the context
        SpecialFieldNotFoundError
            If the special field does not exist.
        """
        logger.fine("Rendering a special field [%s]", field)
        return context.curr_iteration().get_special_field(field)


class MimeoRenderer:
    """A Facade class rendering Mimeo Utils, Vars and Special Fields.

    It contains only class methods.

    Methods
    -------
    render(value: Any) -> Any
        Render a value.
    get_special_field_name(wrapped_field_name: str) -> str
        Extract a special field name.
    is_special_field(special_field: str) -> bool
        Verify if the field is special (of form {:FIELD_NAME:}).
    is_raw_mimeo_util(value: str) -> bool
        Verify if the value is a raw Mimeo Util.
    is_parametrized_mimeo_util(value: dict)
        Verify if the value is a parametrized Mimeo Util.
    """

    _VARS_PATTERN: Pattern = re.compile(".*({[A-Z_0-9]+})")
    _SPECIAL_FIELDS_PATTERN: Pattern = re.compile(".*({:([a-zA-Z]+:)?[-_a-zA-Z0-9]+:})")

    @classmethod
    def get_special_field_name(
            cls,
            wrapped_field_name: str,
    ) -> str:
        """Extract a special field name.

        Parameters
        ----------
        wrapped_field_name : str
            A field name wrapped with curly braces and colons,
            e.g. :Field:

        Returns
        -------
        str
            A special field name

        Raises
        ------
        NotASpecialFieldError
            If the `wrapped_field_name` is not of form :FIELD_NAME:
        """
        if not cls.is_special_field(wrapped_field_name):
            raise NotASpecialFieldError(wrapped_field_name)

        return wrapped_field_name[1:][:-1]

    @classmethod
    def is_special_field(
            cls,
            special_field: str,
    ) -> bool:
        """Verify if the field is special (of form :FIELD_NAME:).

        Parameters
        ----------
        special_field : str
            A special field

        Returns
        -------
        bool
            True if the field is of form :FIELD_NAME:. Otherwise,
            False.
        """
        return bool(re.match(r"^:([a-zA-Z]+:)?[-_a-zA-Z0-9]+:$", special_field))

    @classmethod
    def is_raw_mimeo_util(
            cls,
            value: str,
    ) -> bool:
        """Verify if the value is a raw Mimeo Util.

        Parameters
        ----------
        value : str
            A string value

        Returns
        -------
        bool
            True if the value is a raw Mimeo Util, e.g. {random_str}.
            Otherwise, False.
        """
        raw_mimeo_utils = UtilsRenderer.MIMEO_UTILS.keys()
        raw_mimeo_utils_re = "^{(" + "|".join(raw_mimeo_utils) + ")}$"
        return bool(re.match(raw_mimeo_utils_re, value))

    @classmethod
    def is_parametrized_mimeo_util(
            cls,
            value: dict,
    ):
        """Verify if the value is a parametrized Mimeo Util.

        Parameters
        ----------
        value : dict
            A dict value

        Returns
        -------
        bool
            True if the value is a dictionary having only one key,
            "_mimeo_util". Otherwise, False.
        """
        return (isinstance(value, dict) and
                len(value) == 1 and
                cc.MODEL_MIMEO_UTIL_KEY in value)

    @classmethod
    def is_reference(
            cls,
            value: str,
    ) -> bool:
        """Verify if the value is a Mimeo Reference.

        Parameters
        ----------
        value : str
            A string value

        Returns
        -------
        bool
            True if the value is a Mimeo Reference. Otherwise, False.
        """
        reference_names = MimeoContextManager().get_ref_names()
        if len(reference_names) == 0:
            return False
        reference_re = "^{(" + "|".join(reference_names) + ")}$"
        return bool(re.match(reference_re, value))

    @classmethod
    def render(
            cls,
            value: Any,
    ) -> Any:
        """Render a value.

        This method renders a value accordingly to its type and form. If the value
        takes a form of Mimeo Util it is rendered as a Mimeo Util (raw or parametrized);
        if it takes a form of a special field this renderer will try to reach it from
        the current context; when the value takes a form of a Mimeo Var, then it uses
        Mimeo Vars defined in Mimeo Config; when it is a Mimeo Ref the renderer will
        try to find a reference value using its metadata from Mimeo Configuration.
        Otherwise, the raw value is returned.
        It is recursively called to return a final value.

        Parameters
        ----------
        value : Any
            A value to be rendered

        Returns
        -------
        Any
            A rendered value

        Raises
        ------
        InstanceNotAliveError
            If the MimeoContextManager instance is not alive
        UninitializedContextIterationError
            If no iteration has been initialized yet for the context
        VarNotFoundError
            If the Mimeo Var does not exist
        InvalidMimeoUtilError
            If the Mimeo Util node has missing _name property, or it
            does not match any Mimeo Util.
        InvalidValueError
            If Mimeo Util node is incorrectly parametrized
        OutOfStockError
            If all unique values have been consumed already
        DataNotFoundError
            If database does not contain the expected value
        InvalidSexError
            If the First Name Mimeo Util has not supported `sex`
            parameter value assigned.
        ReferenceNotFoundError
            If there's such a reference configured
        NonPopulatedReferenceError
            If the reference has no values
        NoCorrespondingReferenceError
            If there was no value cached in a corresponding iteration of the source
            context
        """
        logger.fine("Rendering a value [%s]", value)
        try:
            if isinstance(value, str):
                value = cls._render_string_value(value)
            if cls.is_parametrized_mimeo_util(value):
                value = cls._render_parametrized_mimeo_util(value)
        except Exception:
            logger.exception("An error occurred for [%s].", value)
            raise
        return value

    @classmethod
    def _render_string_value(
            cls,
            value: str,
    ) -> Any:
        """Render a string value.

        Depending on value form it can render it as a raw value, a special field,
        a Mimeo Var, a raw Mimeo Util or Mimeo Ref.

        Parameters
        ----------
        value : str
            A string value

        Returns
        -------
        Any
            A rendered value
        """
        if cls._SPECIAL_FIELDS_PATTERN.match(value):
            return cls._render_special_field(value)
        if cls._VARS_PATTERN.match(value):
            return cls._render_var(value)
        if cls.is_raw_mimeo_util(value):
            return cls._render_raw_mimeo_util(value)
        if cls.is_reference(value):
            return cls._render_reference(value)
        return value

    @classmethod
    def _render_special_field(
            cls,
            value: str,
    ) -> Any:
        """Render a value containing a Mimeo Special Field.

        This method finds first special field and replaces all
        occurrences. Then the result is passed to the render() method
        again, to return a final value.

        Parameters
        ----------
        value : str
            A value containing a Mimeo Special Field

        Returns
        -------
        Any
            A rendered value

        Raises
        ------
        UninitializedContextIterationError
            If no iteration has been initialized yet for the context
        SpecialFieldNotFoundError
            If the special field does not exist.
        """
        match = next(cls._SPECIAL_FIELDS_PATTERN.finditer(value))
        wrapped_special_field = match.group(1)
        r_val = SpecialFieldsRenderer.render(wrapped_special_field[2:][:-2])
        logger.fine("Rendered special field value [%s]", r_val)
        if len(wrapped_special_field) != len(value):
            r_val = str(r_val).lower() if isinstance(r_val, bool) else str(r_val)
            r_val = value.replace(wrapped_special_field, str(r_val))
        return cls.render(r_val)

    @classmethod
    def _render_var(
            cls,
            value: str,
    ) -> Any:
        """Render a value containing a Mimeo Var.

        This method finds first variable and replaces all occurrences.
        Then the result is passed to the render() method again, to
        return a final value.

        Parameters
        ----------
        value : str
            A value containing a Mimeo Var

        Returns
        -------
        Any
            A rendered value

        Raises
        ------
        InstanceNotAliveError
            If the MimeoContextManager instance is not alive
        VarNotFoundError
            If the Mimeo Var with the `var` provided does not exist
        """
        match = next(cls._VARS_PATTERN.finditer(value))
        wrapped_var = match.group(1)
        r_val = VarsRenderer.render(wrapped_var[1:][:-1])
        logger.fine("Rendered variable value [%s]", r_val)
        if cls.is_parametrized_mimeo_util(r_val):
            r_val = cls._render_parametrized_mimeo_util(r_val)
        if len(wrapped_var) != len(value):
            r_val = str(r_val).lower() if isinstance(r_val, bool) else str(r_val)
            r_val = value.replace(wrapped_var, str(r_val))
        return cls.render(r_val)

    @classmethod
    def _render_reference(
            cls,
            value: str,
    ) -> Any:
        """Render a Mimeo Ref.

        Parameters
        ----------
        value : str
            A Mimeo Ref

        Returns
        -------
        Any
            A rendered value

        Raises
        ------
        InstanceNotAliveError
            If the MimeoContextManager instance is not alive
        ReferenceNotFoundError
            If there's such a reference configured
        NonPopulatedReferenceError
            If the reference has no values
        NoCorrespondingReferenceError
            If there was no value cached in a corresponding iteration of the source
            context
        """
        rendered_value = RefsRenderer.render(value[1:][:-1])
        return cls.render(rendered_value)

    @classmethod
    def _render_raw_mimeo_util(
            cls,
            value: str,
    ) -> Any:
        """Render a raw Mimeo Util.

        Parameters
        ----------
        value : str
            A raw Mimeo Util

        Returns
        -------
        Any
            A rendered value

        Raises
        ------
        OutOfStockError
            If all unique values have been consumed already
        DataNotFoundError
            If database does not contain the expected value
        """
        rendered_value = UtilsRenderer.render_raw(value[1:][:-1])
        return cls.render(rendered_value)

    @classmethod
    def _render_parametrized_mimeo_util(
            cls,
            value: dict,
    ) -> Any:
        """Render a parametrized Mimeo Util.

        Before the Mimeo Util itself will be rendered, first all its
        parameters (except name) are rendered.

        Parameters
        ----------
        value : str
            A parametrized Mimeo Util

        Returns
        -------
        Any
            A rendered value

        Raises
        ------
        InvalidValueError
            If a Mimeo Util is incorrectly parametrized.
        OutOfStockError
            If all unique values have been consumed already
        DataNotFoundError
            If database does not contain the expected value
        InvalidSexError
            If the First Name Mimeo Util has not supported `sex`
            parameter value assigned.
        """
        mimeo_util = value[cc.MODEL_MIMEO_UTIL_KEY]
        mimeo_util = cls._render_mimeo_util_parameters(mimeo_util)
        logger.fine("Pre-rendered mimeo util [%s]", mimeo_util)
        r_val = UtilsRenderer.render_parametrized(mimeo_util)
        return cls.render(r_val)

    @classmethod
    def _render_mimeo_util_parameters(
            cls,
            mimeo_util_config: dict,
    ) -> dict:
        """Render Mimeo Util's parameters.

        Parameters
        ----------
        mimeo_util_config : dict
            A parametrized Mimeo Util

        Returns
        -------
        dict
            A Mimeo Util with pre-rendered parameters

        Raises
        ------
        InvalidValueError
            If a Mimeo Util is incorrectly parametrized.
        OutOfStockError
            If all unique values have been consumed already
        DataNotFoundError
            If database does not contain the expected value
        InvalidSexError
            If the First Name Mimeo Util has not supported `sex`
            parameter value assigned.
        """
        logger.fine("Rendering mimeo util parameters")
        return {key: cls.render(value) if key != "_name" else value
                for key, value in mimeo_util_config.items()}



"""The Mimeo Logging package.

It contains helpful modules to customize logs produced by Mimeo:
* filters
    The Mimeo Logging Filters module.

This package exports the following classes:
* RegularFilter:
    A Filter subclass filtering out log records below INFO level.
* DetailedFilter:
    A Filter subclass filtering out log records above DEBUG level.

It also includes a single function to set up logging in the application:
* setup_logging
    Set up customized logging configuration.

To use this package, simply import the desired class or function:
    from mimeo.logging import RegularFilter
    from mimeo.logging import setup_logging
"""
from __future__ import annotations

import logging.config

import yaml
from haggis.logs import add_logging_level

from mimeo import tools

from .filters import DetailedFilter, RegularFilter

__all__ = ["DetailedFilter", "RegularFilter", "setup_logging"]


def setup_logging():
    """Set up customized logging configuration.

    The configuration is stored in the logging.yaml file in
    the resources package. Besides that a new log level is defined,
    FINE, together with a method logger.fine().
    """
    add_logging_level("FINE", logging.DEBUG - 1)
    with tools.get_resource("logging.yaml") as config_file:
        config = yaml.safe_load(config_file.read())
        logging.config.dictConfig(config)



"""The Mimeo Utils module.

This module contains all Mimeo Utils. It exports the following classes:
    * MimeoUtil
        A superclass for all Mimeo Utils.
    * RandomStringUtil
        A MimeoUtil implementation rendering a random string value.
    * RandomIntegerUtil
        A MimeoUtil implementation rendering a random integer value.
    * RandomItemUtil
        A MimeoUtil implementation rendering a random item.
    * DateUtil
        A MimeoUtil implementation rendering a stringified date value.
    * DateTimeUtil
        A MimeoUtil implementation rendering a stringified date time value.
    * AutoIncrementUtil
        A MimeoUtil implementation rendering an auto incremented ID.
    * CurrentIterationUtil
        A MimeoUtil implementation rendering a current iteration ID.
    * KeyUtil
        A MimeoUtil implementation rendering a unique identifier.
    * CityUtil
        A MimeoUtil implementation rendering city names.
    * CountryUtil
        A MimeoUtil implementation rendering country details.
    * CurrencyUtil
        A MimeoUtil implementation rendering currency details.
    * FirstNameUtil
        A MimeoUtil implementation rendering forenames.
    * LastNameUtil
        A MimeoUtil implementation rendering surnames.
"""
from __future__ import annotations

import random
import string
from abc import ABCMeta, abstractmethod
from datetime import date, datetime, timedelta
from typing import Any

from mimeo.context import MimeoContext, MimeoContextManager
from mimeo.context.decorators import mimeo_context
from mimeo.database import Country, Currency, MimeoDB
from mimeo.database.exc import DataNotFoundError, InvalidSexError
from mimeo.utils.exc import InvalidValueError


class MimeoUtil(metaclass=ABCMeta):
    """A superclass for all Mimeo Utils.

    It defines abstract methods to be implemented in each subclass.

    Methods
    -------
    render
        Render a value.
    """

    @classmethod
    def __subclasshook__(
            cls,
            subclass: MimeoUtil,
    ):
        """Verify if a subclass implements all abstract methods.

        Parameters
        ----------
        subclass : MimeoUtil
            A MimeoUtil subclass

        Returns
        -------
        bool
            True if the subclass includes the render method and KEY
            attribute
        """
        return ("KEY" in subclass.__dict__ and
                not callable(subclass.KEY) and
                "render" in subclass.__dict__ and
                callable(subclass.render))

    @abstractmethod
    def render(
            self,
    ) -> Any:
        """Render a value.

        It is an abstract method to implement in subclasses
        """
        raise NotImplementedError


class RandomStringUtil(MimeoUtil):
    """A MimeoUtil implementation rendering a random string value.

    Methods
    -------
    render
        Render a random string value.

    Attributes
    ----------
    KEY : str
        A Mimeo Util key
    """

    KEY: str = "random_str"

    def __init__(
            self,
            length: int = 20,
            **kwargs,
    ):
        """Initialize RandomStringUtil class.

        Parameters
        ----------
        length : int, default 20
            A length of a string to render
        kwargs : dict
            Arbitrary keyword arguments (ignored).
        """
        self._length: int = length

    def render(
            self,
    ) -> str:
        """Render a random string value.

        Returns
        -------
        str
            A random string value

        Raises
        ------
        InvalidValueError
            If the length param is negative
        """
        if self._length < 0:
            raise InvalidValueError(InvalidValueError.Code.ERR_1,
                                    util=self.KEY,
                                    length=self._length)
        return "".join(random.choice(string.ascii_letters) for _ in range(self._length))


class RandomIntegerUtil(MimeoUtil):
    """A MimeoUtil implementation rendering a random integer value.

    Methods
    -------
    render
        Render a random integer value.

    Attributes
    ----------
    KEY : str
        A Mimeo Util key
    """

    KEY: str = "random_int"

    def __init__(
            self,
            start: int = 1,
            limit: int = 100,
            **kwargs,
    ):
        """Initialize RandomIntegerUtil class.

        Parameters
        ----------
        start : int, default 1
            A lower bound for integers (inclusive)
        limit : int, default 100
            An upper bound for integers (inclusive)
        kwargs : dict
            Arbitrary keyword arguments (ignored)
        """
        self._start: int = start
        self._limit: int = limit

    def render(
            self,
    ) -> int:
        """Render a random integer value.

        Returns
        -------
        int
            A random integer value

        Raises
        ------
        InvalidValueError
            If the limit param is lower than start
        """
        if self._start > self._limit:
            raise InvalidValueError(InvalidValueError.Code.ERR_2,
                                    util=self.KEY,
                                    limit=self._limit,
                                    start=self._start)
        return random.randrange(self._start, self._limit + 1)


class RandomItemUtil(MimeoUtil):
    """A MimeoUtil implementation rendering a random item.

    Methods
    -------
    render
        Render a random item.

    Attributes
    ----------
    KEY : str
        A Mimeo Util key
    """

    KEY: str = "random_item"

    def __init__(
            self,
            items: list | None = None,
            **kwargs,
    ):
        """Initialize RandomItemUtil class.

        Parameters
        ----------
        items : int, default ['']
            A list of items from which a value will be picked up
        kwargs : dict
            Arbitrary keyword arguments (ignored)
        """
        self._items: list = items if items is not None and len(items) != 0 else [""]

    def render(
            self,
    ) -> Any:
        """Render a random item.

        Returns
        -------
        Any
            A random item
        """
        length = len(self._items)
        return self._items[random.randrange(0, length)]


class DateUtil(MimeoUtil):
    """A MimeoUtil implementation rendering a stringified date value.

    Methods
    -------
    render
        Render a stringified date value.

    Attributes
    ----------
    KEY : str
        A Mimeo Util key
    """

    KEY: str = "date"

    def __init__(
            self,
            days_delta: int = 0,
            **kwargs,
    ):
        """Initialize DateUtil class.

        Parameters
        ----------
        days_delta : int, default 0
            An integer value of days to add or subtract from today
        kwargs : dict
            Arbitrary keyword arguments (ignored)
        """
        self._days_delta: int = days_delta

    def render(
            self,
    ) -> str:
        """Render a stringified date value.

        Returns
        -------
        str
            A stringified date value in format %Y-%m-%d
        """
        if self._days_delta == 0:
            date_value = date.today()
        else:
            date_value = date.today() + timedelta(days=self._days_delta)
        return date_value.strftime("%Y-%m-%d")


class DateTimeUtil(MimeoUtil):
    """A MimeoUtil implementation rendering a stringified date time value.

    Methods
    -------
    render
        Render a stringified date time value.

    Attributes
    ----------
    KEY : str
        A Mimeo Util key
    """

    KEY: str = "date_time"

    def __init__(self,
                 days_delta: int = 0,
                 hours_delta: int = 0,
                 minutes_delta: int = 0,
                 seconds_delta: int = 0,
                 **kwargs):
        """Initialize DateTimeUtil class.

        Parameters
        ----------
        days_delta : int, default 0
            An integer value of days to add or subtract from now
        hours_delta : int, default 0
            An integer value of hours to add or subtract from now
        minutes_delta : int, default 0
            An integer value of minutes to add or subtract from now
        seconds_delta : int, default 0
            An integer value of seconds to add or subtract from now
        kwargs : dict
            Arbitrary keyword arguments (ignored)
        """
        self._days_delta: int = days_delta
        self._hours_delta: int = hours_delta
        self._minutes_delta: int = minutes_delta
        self._seconds_delta: int = seconds_delta

    def render(
            self,
    ) -> str:
        """Render a stringified date time value.

        Returns
        -------
        str
            A stringified date time value in format %Y-%m-%dT%H:%M:%S
        """
        time_value = datetime.now() + timedelta(days=self._days_delta,
                                                hours=self._hours_delta,
                                                minutes=self._minutes_delta,
                                                seconds=self._seconds_delta)
        return time_value.strftime("%Y-%m-%dT%H:%M:%S")


class AutoIncrementUtil(MimeoUtil):
    """A MimeoUtil implementation rendering an auto incremented ID.

    It is a Mimeo Context-dependent Mimeo Util. Rendered ID value is
    pulled from the Context.

    Methods
    -------
    render
        Render an auto incremented identifier.

    Attributes
    ----------
    KEY : str
        A Mimeo Util key
    """

    KEY: str = "auto_increment"

    def __init__(
            self,
            pattern: str = "{:05d}",
            **kwargs,
    ):
        """Initialize AutoIncrementUtil class.

        Parameters
        ----------
        pattern : str, default '{:05d}'
            A pattern to inject incremented ID
        kwargs : dict
            Arbitrary keyword arguments (ignored)
        """
        self._pattern: str = pattern

    @mimeo_context
    def render(
            self,
            context: MimeoContext | None = None,
    ) -> str:
        """Render an auto incremented identifier.

        Parameters
        ----------
        context : MimeoContext, default None
            A current Mimeo Context injected by a decorator

        Returns
        -------
        str
            An auto incremented identifier in a parametrized format

        Raises
        ------
        InvalidValueError
            If the pattern is not a string value
        """
        try:
            identifier = context.next_id()
            return self._pattern.format(identifier)
        except AttributeError:
            context.prev_id()
            raise InvalidValueError(InvalidValueError.Code.ERR_3,
                                    util=self.KEY,
                                    type="string",
                                    param_name="pattern",
                                    param_val=self._pattern) from AttributeError


class CurrentIterationUtil(MimeoUtil):
    """A MimeoUtil implementation rendering a current iteration ID.

    It is a Mimeo Context-dependent Mimeo Util. Rendered ID value is
    pulled from the Context.

    Methods
    -------
    render
        Render a current iteration ID.

    Attributes
    ----------
    KEY : str
        A Mimeo Util key
    """

    KEY: str = "curr_iter"

    def __init__(
            self,
            context: str | None = None,
            **kwargs,
    ):
        """Initialize CurrentIterationUtil class.

        Parameters
        ----------
        context : str, default None
            A context name to reach the current iteration
        kwargs : dict
            Arbitrary keyword arguments (ignored)
        """
        self._context_name: str = context

    @mimeo_context
    def render(
            self,
            context: MimeoContext | None = None,
    ) -> int:
        """Render a current iteration ID.

        Parameters
        ----------
        context : MimeoContext, default None
            A current Mimeo Context injected by a decorator

        Returns
        -------
        int
            A specific Mimeo Context's current iteration ID
        """
        if self._context_name is not None:
            context = MimeoContextManager().get_context(self._context_name)
        return context.curr_iteration().id


class KeyUtil(MimeoUtil):
    """A MimeoUtil implementation rendering a unique identifier.

    It is a Mimeo Context-dependent Mimeo Util. Rendered identifiers
    are stored in a MimeoContext and pulled from it when the Mimeo
    Util is parametrized to do so.

    Methods
    -------
    render
        Render a unique identifier.

    Attributes
    ----------
    KEY : str
        A Mimeo Util key
    """

    KEY: str = "key"

    def __init__(
            self,
            context: str | None = None,
            iteration: int | None = None,
            **kwargs,
    ):
        """Initialize KeyUtil class.

        Parameters
        ----------
        context : str, default None
            A context name to reach already generated identifier
        iteration : int, default None
            A iteration id to reach already generated identifier
        kwargs : dict
            Arbitrary keyword arguments (ignored)
        """
        self._context_name: str = context
        self._iteration: int = iteration

    @mimeo_context
    def render(
            self,
            context: MimeoContext | None = None,
    ) -> str:
        """Render a unique identifier.

        By default, Key Mimeo Util renders identifier from the current
        iteration of the current Mimeo Context.
        If the context name is parametrized and iteration is not, then
        identifier is pulled from the current iteration of THIS
        context. If the iteration is parametrized, then the identifier
        is pulled from the specific iteration of Mimeo Context (current
        or parametrized).

        Parameters
        ----------
        context : MimeoContext, default None
            A current Mimeo Context injected by a decorator

        Returns
        -------
        str
            A unique identifier
        """
        if self._context_name is not None:
            context = MimeoContextManager().get_context(self._context_name)
        if self._iteration is None:
            iteration = context.curr_iteration()
        else:
            iteration = context.get_iteration(self._iteration)
        return iteration.key


class CityUtil(MimeoUtil):
    """A MimeoUtil implementation rendering city names.

    It is a Mimeo Context-dependent Mimeo Util only when parametrized
    to generate unique city names.

    Methods
    -------
    render
        Render a city name.

    Attributes
    ----------
    KEY : str
        A Mimeo Util key
    """

    KEY: str = "city"
    _MIMEO_DB: MimeoDB = MimeoDB()

    def __init__(
            self,
            unique: bool = True,
            country: str | None = None,
            **kwargs,
    ):
        """Initialize CityUtil class.

        Parameters
        ----------
        unique : bool, default True
            Indicates if rendered city names will be unique across
            a Mimeo Context
        country : str, default None
            A country limiting city names that can be rendered
        kwargs : dict
            Arbitrary keyword arguments (ignored)
        """
        self._unique: bool = unique
        self._country: str = country

    @mimeo_context
    def render(
            self,
            context: MimeoContext | None = None,
    ) -> str:
        """Render a city name.

        By default, City Mimeo Util generates a unique city name across
        a Mimeo Context without `country` limitation. If `country` is
        parametrized, then this Mimeo Util will render only those
        city names that lie in the specific country.

        Parameters
        ----------
        context : MimeoContext, default None
            A current Mimeo Context injected by a decorator

        Returns
        -------
        str
            A city name

        Raises
        ------
        OutOfStockError
            If all unique cities have been consumed already
        DataNotFoundError
            If database does not contain any cities for the provided
            `country`
        """
        if self._country is None:
            if self._unique:
                index = context.next_city_index()
            else:
                index = random.randrange(MimeoDB.NUM_OF_CITIES)
            city = self._MIMEO_DB.get_city_at(index)
        else:
            country_cities = self._MIMEO_DB.get_cities_of(self._country)
            country_cities_count = len(country_cities)
            if country_cities_count == 0:
                raise DataNotFoundError(DataNotFoundError.Code.ERR_2,
                                        data="city",
                                        param_name="country",
                                        param_val=self._country)

            if self._unique:
                index = context.next_city_index(self._country)
            else:
                index = random.randrange(country_cities_count)
            city = country_cities[index]

        return city.name_ascii


class CountryUtil(MimeoUtil):
    """A MimeoUtil implementation rendering country details.

    It is a Mimeo Context-dependent Mimeo Util only when parametrized
    to generate unique country names.

    Methods
    -------
    render
        Render a country detail (name, ISO3 or ISO2 code).

    Attributes
    ----------
    KEY : str
        A Mimeo Util key
    """

    KEY: str = "country"

    _VALUE_NAME: str = "name"
    _VALUE_ISO3: str = "iso3"
    _VALUE_ISO2: str = "iso2"
    _SUPPORTED_VALUES: tuple = (_VALUE_NAME, _VALUE_ISO3, _VALUE_ISO2)
    _MIMEO_DB: MimeoDB = MimeoDB()

    def __init__(
            self,
            value: str | None = None,
            unique: bool = True,
            country: str | None = None,
            **kwargs,
    ):
        """Initialize CountryUtil class.

        Parameters
        ----------
        value : str, default 'name'
            Indicates which country detail should be rendered:
            name, ISO3 code or ISO2 code.
        unique : bool, default True
            Indicates if rendered country names will be unique across
            a Mimeo Context
        country : str, default None
            A one country detail to get its other detail
        kwargs : dict
            Arbitrary keyword arguments (ignored)
        """
        self._value: str = value if value is not None else self._VALUE_NAME
        self._unique: bool = unique
        self._country: str = country

    @mimeo_context
    def render(
            self,
            context: MimeoContext | None = None,
    ) -> str:
        """Render a country name.

        By default, Country Mimeo Util generates a unique country name
        across a Mimeo Context. If `value` is parametrized, then it
        will render a specific country detail (name, ISO3 code or ISO2
        code). This Mimeo Util allows you to provide a `country` detail
        (e.g. ISO3 code) to get another detail (e.g. name).
        When `country` is parametrized, then the unique parameter is
        ignored.

        Parameters
        ----------
        context : MimeoContext, default None
            A current Mimeo Context injected by a decorator

        Returns
        -------
        str
            A country detail

        Raises
        ------
        InvalidValueError
            If the `value` parameter value is not supported
        OutOfStockError
            If all unique countries have been consumed already
        DataNotFoundError
            If database does not contain the provided `country`
        """
        if self._value == self._VALUE_NAME:
            return self._get_country(context).name
        if self._value == self._VALUE_ISO3:
            return self._get_country(context).iso_3
        if self._value == self._VALUE_ISO2:
            return self._get_country(context).iso_2
        raise InvalidValueError(InvalidValueError.Code.ERR_4,
                                util=self.KEY,
                                value=self._value,
                                supported_values=self._SUPPORTED_VALUES)

    def _get_country(
            self,
            context: MimeoContext,
    ) -> Country:
        if self._country is not None:
            countries = self._MIMEO_DB.get_countries()
            country_found = next(
                filter(lambda c: self._country in [c.name, c.iso_3, c.iso_2],
                       countries),
                None,
            )
            if country_found is None:
                raise DataNotFoundError(DataNotFoundError.Code.ERR_1,
                                        data="country",
                                        value=self._country)
            return country_found

        if self._unique:
            index = context.next_country_index()
        else:
            index = random.randrange(MimeoDB.NUM_OF_COUNTRIES)
        return self._MIMEO_DB.get_country_at(index)


class CurrencyUtil(MimeoUtil):
    """A MimeoUtil implementation rendering currency details.

    It is a Mimeo Context-dependent Mimeo Util only when parametrized
    to generate unique currency codes.

    Methods
    -------
    render
        Render a currency detail (code or name).

    Attributes
    ----------
    KEY : str
        A Mimeo Util key
    """

    KEY: str = "currency"

    _VALUE_CODE: str = "code"
    _VALUE_NAME: str = "name"
    _SUPPORTED_VALUES: tuple = (_VALUE_CODE, _VALUE_NAME)
    _MIMEO_DB: MimeoDB = MimeoDB()

    def __init__(
            self,
            value: str | None = None,
            unique: bool = False,
            country: str | None = None,
            **kwargs,
    ):
        """Initialize CurrencyUtil class.

        Parameters
        ----------
        value : str, default 'code'
            Indicates which currency detail should be rendered:
            code or name.
        unique : bool, default False
            Indicates if rendered currency codes will be unique across
            a Mimeo Context
        country : str, default None
            A country of which the currency should be rendered
        kwargs : dict
            Arbitrary keyword arguments (ignored)
        """
        self._value: str = value if value is not None else self._VALUE_CODE
        self._unique: bool = unique
        self._country: str = country

    @mimeo_context
    def render(
            self,
            context: MimeoContext | None = None,
    ) -> str:
        """Render a currency code.

        By default, Currency Mimeo Util generates a non-unique currency code across
        a Mimeo Context. If `value` is parametrized, then it will render a specific
        currency detail (code or name). This Mimeo Util allows you to provide
        a `country` to get a currency being used there. When `country` is parametrized,
        then the 'unique' parameter is ignored.

        Parameters
        ----------
        context : MimeoContext, default None
            A current Mimeo Context injected by a decorator

        Returns
        -------
        str
            A currency code

        Raises
        ------
        InvalidValueError
            If the `value` parameter value is not supported
        OutOfStockError
            If all unique currencies have been consumed already
        DataNotFoundError
            If database does not contain a currency of the provided `country`
        """
        if self._value == self._VALUE_CODE:
            return self._get_currency(context).code
        if self._value == self._VALUE_NAME:
            return self._get_currency(context).name
        raise InvalidValueError(InvalidValueError.Code.ERR_4,
                                util=self.KEY,
                                value=self._value,
                                supported_values=self._SUPPORTED_VALUES)

    def _get_currency(
            self,
            context: MimeoContext,
    ) -> Currency:
        if self._country is None:
            if self._unique:
                index = context.next_currency_index()
            else:
                index = random.randrange(MimeoDB.NUM_OF_CURRENCIES)
            currency = self._MIMEO_DB.get_currency_at(index)
        else:
            currency = self._MIMEO_DB.get_currency_of(self._country)
            if currency is None:
                raise DataNotFoundError(DataNotFoundError.Code.ERR_2,
                                        data="currency",
                                        param_name="country",
                                        param_val=self._country)
        return currency


class FirstNameUtil(MimeoUtil):
    """A MimeoUtil implementation rendering forenames.

    It is a Mimeo Context-dependent Mimeo Util only when parametrized
    to generate unique forenames.

    Methods
    -------
    render
        Render a first name.

    Attributes
    ----------
    KEY : str
        A Mimeo Util key
    """

    KEY: str = "first_name"
    __MIMEO_DB: MimeoDB = MimeoDB()

    def __init__(
            self,
            unique: bool = True,
            sex: str | None = None,
            **kwargs,
    ):
        """Initialize FirstNameUtil class.

        Parameters
        ----------
        unique : bool, default True
            Indicates if rendered forenames will be unique across
            a Mimeo Context
        sex : str, default None
            A sex to limit forenames
        kwargs : dict
            Arbitrary keyword arguments (ignored)
        """
        self._unique: bool = unique
        self._sex: str = self._standardize_sex(sex)

    @mimeo_context
    def render(
            self,
            context: MimeoContext | None = None,
    ) -> str:
        """Render a first name.

        By default, First Name Mimeo Util generates a unique forename
        across a Mimeo Context without sex limitation. If `sex` is
        parametrized, then this Mimeo Util will render only those
        first names that are assigned to the specific sex.

        Parameters
        ----------
        context : MimeoContext, default None
            A current Mimeo Context injected by a decorator

        Returns
        -------
        str
            A first name

        Raises
        ------
        OutOfStockError
            If all unique first names have been consumed already
        InvalidSexError
            If the `sex` parameter value is not supported
        """
        if self._sex is None:
            if self._unique:
                index = context.next_first_name_index()
            else:
                index = random.randrange(MimeoDB.NUM_OF_FIRST_NAMES)
            first_name = self.__MIMEO_DB.get_first_name_at(index)
        else:
            first_name_for_sex = self.__MIMEO_DB.get_first_names_by_sex(self._sex)
            first_name_for_sex_count = len(first_name_for_sex)

            if self._unique:
                index = context.next_first_name_index(self._sex)
            else:
                index = random.randrange(first_name_for_sex_count)
            first_name = first_name_for_sex[index]

        return first_name.name

    @classmethod
    def _standardize_sex(
            cls,
            sex: str,
    ):
        if sex is None:
            return sex
        if sex.upper() in ["M", "MALE"]:
            return "M"
        if sex.upper() in ["F", "FEMALE"]:
            return "F"
        raise InvalidSexError(("M", "F", "Male", "Female"))


class LastNameUtil(MimeoUtil):
    """A MimeoUtil implementation rendering surnames.

    It is a Mimeo Context-dependent Mimeo Util only when parametrized
    to generate unique surnames.

    Methods
    -------
    render
        Render a last name.

    Attributes
    ----------
    KEY : str
        A Mimeo Util key
    """

    KEY: str = "last_name"
    __MIMEO_DB: MimeoDB = MimeoDB()

    def __init__(
            self,
            unique: bool = True,
            **kwargs,
    ):
        """Initialize LastNameUtil class.

        Parameters
        ----------
        unique : bool, default True
            Indicates if rendered forenames will be unique across
            a Mimeo Context
        kwargs : dict
            Arbitrary keyword arguments (ignored)
        """
        self._unique: bool = unique

    @mimeo_context
    def render(
            self,
            context: MimeoContext | None = None,
    ) -> str:
        """Render a last name.

        By default, First Name Mimeo Util generates a unique surname
        across a Mimeo Context.

        Parameters
        ----------
        context : MimeoContext, default None
            A current Mimeo Context injected by a decorator

        Returns
        -------
        str
            A last name

        Raises
        ------
        OutOfStockError
            If all unique last names have been consumed already
        """
        if self._unique:
            index = context.next_first_name_index()
        else:
            index = random.randrange(MimeoDB.NUM_OF_FIRST_NAMES)
        return self.__MIMEO_DB.get_last_name_at(index)



"""The Mimeo CLI package.

It contains the following modules:
* job
    The Mimeo Job module.
* parsers
    The Mimeo Parser module.
* exc
    The Mimeo CLI Exceptions module.

The Mimeo CLI package exports classes responsible for preparation and executing a Mimeo
processing job from the command line:
* MimeoJob
    A class representing a single Mimeo processing job.
* MimeoArgumentParser
    A custom ArgumentParser for the Mimeo CLI.
* MimeoConfigParser
    A class parsing source Mimeo Configuration with Mimeo arguments.

To use this package, simply import it:
    from mimeo.cli import MimeoJob
"""
from .parsers import MimeoArgumentParser, MimeoConfigParser
from .job import MimeoJob

__all__ = ["MimeoJob", "MimeoArgumentParser", "MimeoConfigParser"]



"""The Mimeo Database module.

It exports a class related to all Mimeo CSV data:
    * MimeoDB
        Facade class exposing READ operations on all Mimeo CSV data.
"""
from __future__ import annotations

from mimeo.database import (CitiesDB, City, CountriesDB, Country, CurrenciesDB,
                            Currency, FirstName, FirstNamesDB, LastNamesDB)


class MimeoDB:
    """Facade class exposing READ operations on all Mimeo CSV data.

    It combines all attributes and methods exposed by downstream
    classes:
    - CitiesDB
    - CountriesDB
    - CurrenciesDB
    - FirstNamesDB
    - LastNamesDB

    Attributes
    ----------
    NUM_OF_CITIES : int
        A number of rows in cities CSV data
    NUM_OF_COUNTRIES : int
        A number of rows in countries CSV data
    NUM_OF_CURRENCIES : int
        A number of rows in currencies CSV data
    NUM_OF_FIRST_NAMES : int
        A number of rows in forenames CSV data
    NUM_OF_LAST_NAMES : int
        A number of rows in surnames CSV data

    Methods
    -------
    get_cities() -> list[City]
        Get all cities.
    get_cities_of(country: str) -> list[City]
        Get cities of a specific country.
    get_city_at(index: int) -> City
        Get a city at `index` position.
    get_currencies() -> list[Currency]
        Get all currencies.
    get_currency_at(index: int) -> Currency
        Get a currency at `index` position.
    get_currency_of(country: str) -> Currency | None
        Get a currency at `index` position.
    get_countries() -> list[Country]
        Get all countries.
    get_country_at(index: int) -> Country
        Get a country at `index` position.
    get_country_by_iso_3(iso_3: str) -> Country
        Get a country having a specific ISO3 code.
    get_country_by_iso_3(iso_2: str) -> Country
        Get a country having a specific ISO2 code.
    get_country_by_name(name: str) -> Country
        Get a country having a specific name.
    get_first_names() -> list[FirstName]
        Get all first names.
    get_first_names_by_sex(sex: str) -> list[FirstName]
        Get first names for a specific sex.
    get_first_name_at(index: int) -> FirstName
        Get a first name at `index` position.
    get_last_names() -> list[str]
        Get all last names.
    get_last_name_at(index: int) -> str
        Get a last name at `index` position.
    """

    NUM_OF_CITIES: int = CitiesDB.NUM_OF_RECORDS
    NUM_OF_COUNTRIES: int = CountriesDB.NUM_OF_RECORDS
    NUM_OF_CURRENCIES: int = CurrenciesDB.NUM_OF_RECORDS
    NUM_OF_FIRST_NAMES: int = FirstNamesDB.NUM_OF_RECORDS
    NUM_OF_LAST_NAMES: int = LastNamesDB.NUM_OF_RECORDS

    def __init__(
            self,
    ):
        self._cities_db = CitiesDB()
        self._countries_db = CountriesDB()
        self._currencies_db = CurrenciesDB()
        self._first_names_db = FirstNamesDB()
        self._last_names_db = LastNamesDB()

    def get_cities(
            self,
    ) -> list[City]:
        """Get all cities.

        Returns
        -------
        list[City]
            List of all cities
        """
        return self._cities_db.get_cities()

    def get_cities_of(
            self,
            country: str,
    ) -> list[City]:
        """Get cities of a specific country.

        In contrast to CitiesDB.get_cities_of() method it combines
        CitiesDB and CountriesDB to allow for any country detail
        providing. It can be ISO3 code, ISO2 code and name.
        First it will find the specific country and extract its
        ISO3 code to use in CitiesDB.get_cities_of() call.

        Parameters
        ----------
        country : str
            A country ISO3 / ISO2 code or name to filter cities

        Returns
        -------
        list[City]
            List of cities filtered by country
        """
        countries = filter(
            lambda c: country in [c.iso_3, c.iso_2, c.name],
            self.get_countries())
        country = next(countries, None)
        if country is None:
            return []
        return self._cities_db.get_cities_of(country.iso_3)

    def get_city_at(
            self,
            index: int,
    ) -> City:
        """Get a city at `index` position.

        Parameters
        ----------
        index : int
            A city row index

        Returns
        -------
        City
            A specific city

        Raises
        ------
        InvalidIndexError
            If the provided `index` is out of bounds
        """
        return self._cities_db.get_city_at(index)

    def get_currencies(
            self,
    ) -> list[Currency]:
        """Get all currencies.

        Returns
        -------
        list[Currency]
            List of all currencies
        """
        return self._currencies_db.get_currencies()

    def get_currency_at(
            self,
            index: int,
    ) -> Currency:
        """Get a currency at `index` position.

        Parameters
        ----------
        index : int
            A currency row index

        Returns
        -------
        Currency
            A specific currency

        Raises
        ------
        InvalidIndexError
            If the provided `index` is out of bounds
        """
        return self._currencies_db.get_currency_at(index)

    def get_currency_of(
            self,
            country: str,
    ) -> Currency | None:
        """Get a currency of a specific country.

        In contrast to CurrenciesDB.get_currency_of() method it combines
        CurrenciesDB and CountriesDB to allow for any country detail
        providing. It can be ISO3 code, ISO2 code and name.
        First it will find the specific country and extract its
        name to use in CurrenciesDB.get_currency_of() call.

        Parameters
        ----------
        country : str
            A country ISO3 / ISO2 code or name to filter currency

        Returns
        -------
        Currency
            A currency used in a specific country
        """
        countries = filter(
            lambda c: country in [c.iso_3, c.iso_2, c.name],
            self.get_countries())
        country = next(countries, None)
        if country is None:
            return None
        return self._currencies_db.get_currency_of(country.name)

    def get_countries(
            self,
    ) -> list[Country]:
        """Get all countries.

        Returns
        -------
        list[Country]
            List of all countries
        """
        return self._countries_db.get_countries()

    def get_country_at(
            self,
            index: int,
    ) -> Country:
        """Get a country at `index` position.

        Parameters
        ----------
        index : int
            A country row index

        Returns
        -------
        Country
            A specific country

        Raises
        ------
        InvalidIndexError
            If the provided `index` is out of bounds
        """
        return self._countries_db.get_country_at(index)

    def get_country_by_iso_3(
            self,
            iso_3: str,
    ) -> Country:
        """Get a country having a specific ISO3 code.

        Parameters
        ----------
        iso_3 : str
            An ISO3 code to find a country

        Returns
        -------
        Country
            A specific country or None
        """
        return self._countries_db.get_country_by_iso_3(iso_3)

    def get_country_by_iso_2(
            self,
            iso_2: str,
    ) -> Country:
        """Get a country having a specific ISO2 code.

        Parameters
        ----------
        iso_2 : str
            An ISO2 code to find a country

        Returns
        -------
        Country
            A specific country or None
        """
        return self._countries_db.get_country_by_iso_2(iso_2)

    def get_country_by_name(
            self,
            name: str,
    ) -> Country:
        """Get a country having a specific name.

        Parameters
        ----------
        name : str
            A name to find a country

        Returns
        -------
        Country
            A specific country or None
        """
        return self._countries_db.get_country_by_name(name)

    def get_first_names(
            self,
    ) -> list[FirstName]:
        """Get all first names.

        Returns
        -------
        list[FirstName]
            List of all first names
        """
        return self._first_names_db.get_first_names()

    def get_first_names_by_sex(
            self,
            sex: str,
    ) -> list[FirstName]:
        """Get first names for a specific sex.

        Parameters
        ----------
        sex : str
            A sex value to filter first names

        Returns
        -------
        list[FirstName]
            List of first names filtered by sex

        Raises
        ------
        InvalidSexError
            If the provided `sex` value is not supported
        """
        return self._first_names_db.get_first_names_by_sex(sex)

    def get_first_name_at(
            self,
            index: int,
    ) -> FirstName:
        """Get a first name at `index` position.

        Parameters
        ----------
        index : int
            A first name row index

        Returns
        -------
        FirstName
            A specific first name

        Raises
        ------
        InvalidIndexError
            If the provided `index` is out of bounds
        """
        return self._first_names_db.get_first_name_at(index)

    def get_last_names(
            self,
    ) -> list[str]:
        """Get all last names.

        Returns
        -------
        list[str]
            List of all last names
        """
        return self._last_names_db.get_last_names()

    def get_last_name_at(
            self,
            index: int,
    ) -> str:
        """Get a last name at `index` position.

        Parameters
        ----------
        index : int
            A last name row index

        Returns
        -------
        str
            A last name

        Raises
        ------
        InvalidIndexError
            If the provided `index` is out of bounds
        """
        return self._last_names_db.get_last_name_at(index)



"""The Mimeo Context Manager module.

It exports only one class:
    * MimeoContextManager
        An OnlyOneAlive class managing Mimeo Contexts.
"""
from __future__ import annotations

import random
from types import TracebackType

from mimeo.config import MimeoConfig
from mimeo.config import constants as cc
from mimeo.context import MimeoContext
from mimeo.context.exc import (InvalidReferenceValueError,
                               NoCorrespondingReferenceError,
                               NonPopulatedReferenceError,
                               ReferenceNotFoundError, VarNotFoundError)
from mimeo.meta import Alive, OnlyOneAlive


class MimeoContextManager(Alive, metaclass=OnlyOneAlive):
    """An OnlyOneAlive class managing Mimeo Contexts.

    It allows you to initialize a context, get the currently processing
    context, switch it or reach any other. Besides that it gives you
    access to Mimeo Vars and Mimeo Refs.
    The only way to use it successfully it is by `with` statement:
        with MimeoContextManager(mimeo_config) as mimeo_mng:
            ...

    Methods
    -------
    get_context(self, context: str) -> MimeoContext
        Return a Mimeo Context with a specific name.
    get_current_context(self) -> MimeoContext
        Return the current Mimeo Context.
    set_current_context(self, context: MimeoContext)
        Set the current Mimeo Context.
    get_var(self, variable_name: str) -> str | int | bool | dict
        Return a specific Mimeo Var value.
    cache_ref(self, field_name: str, field_value: str | int | float | bool)
        Cache a field's value in references.
    get_ref(self, ref_name: str) -> str | int | float | bool
        Get a reference value.
    get_ref_names(self) -> list[str]
        Get reference names.
    """

    def __init__(
            self,
            mimeo_config: MimeoConfig | None = None,
    ):
        """Initialize MimeoContextManager class.

        Parameters
        ----------
        mimeo_config : MimeoConfig
            The Mimeo Configuration
        """
        super().__init__()
        self._mimeo_config: MimeoConfig = mimeo_config
        self._vars: dict = {}
        self._refs: dict = {}
        self._contexts: dict = {}
        self._current_context: MimeoContext | None = None

    def __enter__(
            self,
    ) -> MimeoContextManager:
        """Enter the MimeoContextManager instance.

        Extends Alive __enter__ function and initializes vars.

        Returns
        -------
        self : MimeoContextManager
            A MimeoContextManager instance
        """
        super().__enter__()
        self._vars = self._mimeo_config.vars
        self._refs = {ref: [] for ref in self._mimeo_config.refs}
        return self

    def __exit__(
            self,
            exc_type: type | None,
            exc_val: BaseException | None,
            exc_tb: TracebackType | None,
    ) -> None:
        """Exit the MimeoContextManager instance.

        Extends Alive __enter__ function and removes internal
        attributes.

        Parameters
        ----------
        exc_type : type | None
            An exception's type
        exc_val : BaseException | None
            An exception's value
        exc_tb  TracebackType | None
            An exception's traceback

        Returns
        -------
        None
            A None value
        """
        super().__exit__(exc_type, exc_val, exc_tb)
        self._vars = None
        self._contexts = None

    def get_context(
            self,
            context: str,
    ) -> MimeoContext:
        """Return a Mimeo Context with a specific name.

        If the context does not exist, it is initialized.

        Parameters
        ----------
        context : str
            A context's name

        Returns
        -------
        MimeoContext
            A specific Mimeo Context

        Raises
        ------
        InstanceNotAliveError
            If the MimeoContextManager instance is not alive
        """
        super().assert_alive()
        if context not in self._contexts:
            self._contexts[context] = MimeoContext(context)
        return self._contexts[context]

    def get_current_context(
            self,
    ) -> MimeoContext:
        """Return the current Mimeo Context.

        Returns
        -------
        MimeoContext
            The current Mimeo Context

        Raises
        ------
        InstanceNotAliveError
            If the MimeoContextManager instance is not alive
        """
        super().assert_alive()
        return self._current_context

    def set_current_context(
            self,
            context: MimeoContext,
    ):
        """Set the current Mimeo Context.

        Parameters
        ----------
        context : MimeoContext
            A Mimeo Context to be set as the current one

        Raises
        ------
        InstanceNotAliveError
            If the MimeoContextManager instance is not alive
        """
        super().assert_alive()
        self._current_context = context

    def get_var(
            self,
            variable_name: str,
    ) -> str | int | bool | dict:
        """Return a specific Mimeo Var value.

        Parameters
        ----------
        variable_name : str
            The Mimeo Var name

        Returns
        -------
        value : str | int | bool | dict
            The Mimeo Var value

        Raises
        ------
        InstanceNotAliveError
            If the MimeoContextManager instance is not alive
        VarNotFoundError
            If the Mimeo Var with the `variable_name` provided does not
            exist
        """
        super().assert_alive()
        value = self._vars.get(variable_name)
        if value is None:
            raise VarNotFoundError(variable_name)
        return value

    def cache_ref(
            self,
            field_name: str,
            field_value: str | int | float | bool,
    ):
        """Cache a field's value in references.

        It saves the field's value in all references having context configured
        to the current and field's to the one provided. When there's such a reference
        the field's value is not cached.

        Parameters
        ----------
        field_name : str
            A field name
        field_value : str | int | float | bool
            A field value

        Raises
        ------
        InvalidReferenceValueError
            If the field's value is not atomic one
        """
        if isinstance(field_value, (dict, list)):
            raise InvalidReferenceValueError(field_value)
        ref_names = [ref_name
                     for ref_name, ref_meta in self._mimeo_config.refs.items()
                     if ref_meta[cc.REFS_DETAIL_CONTEXT] == self._current_context.name
                     and ref_meta[cc.REFS_DETAIL_FIELD] == field_name]
        for ref_name in ref_names:
            self._refs[ref_name].append(field_value)

    def get_ref(
            self,
            ref_name: str,
    ) -> str | int | float | bool:
        """Get a reference value.

        It provides a reference value depending on its type. For 'any' returns a value
        on any index. For 'parallel' it uses a value produced in a corresponding
        iteration for a source context.

        Parameters
        ----------
        ref_name : str
            A reference name

        Returns
        -------
        str | int | float | bool
            A reference value

        Raises
        ------
        ReferenceNotFoundError
            If there's such a reference configured
        NonPopulatedReferenceError
            If the reference has no values
        NoCorrespondingReferenceError
            If there was no value cached in a corresponding iteration of the source
            context
        """
        ref_meta = self._mimeo_config.refs.get(ref_name)
        if not ref_meta:
            raise ReferenceNotFoundError(ref_name)

        values = self._refs.get(ref_name)
        if len(values) == 0:
            raise NonPopulatedReferenceError(ref_name)

        if ref_meta[cc.REFS_DETAIL_TYPE] == cc.REFS_TYPE_PARALLEL:
            curr_iter = self._current_context.curr_iteration().id
            index = curr_iter - 1
            if index >= len(values):
                raise NoCorrespondingReferenceError(ref_name, curr_iter)
        else:
            index = random.randrange(0, len(values))
        return values[index]

    def get_ref_names(
            self,
    ) -> list[str]:
        """Get reference names.

        Returns
        -------
        list
            List of reference names
        """
        return list(self._mimeo_config.refs.keys())




"""The Mimeo Generators Exceptions module.

It contains all custom exceptions related to content of Mimeo Models:
    * UnsupportedStructureError
        A custom Exception class for an unsupported structure.
"""


from __future__ import annotations


class UnsupportedStructureError(Exception):
    """A custom Exception class for an unsupported structure.

    Raised while attempting to generate XML data from a Mimeo Model having a list
    of non-only atomic / non-only dict items.
    """

    def __init__(
            self,
            field_name: str,
            structure: list,
    ):
        """Initialize UnsupportedStructureError exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        field_name : str
            A field name
        structure : list
            An unsupported structure
        """
        super().__init__("An array can include only atomic types (including Mimeo "
                         "Utils) or only JSON objects (when output format is XML)! "
                         f"Unsupported structure found in {field_name}: {structure}")



"""The Mimeo Generator Factory module.

It exports only one class:
    * GeneratorFactory
        A Factory class instantiating a Generator based on Mimeo Config.
"""
from __future__ import annotations

from mimeo.config import constants as cc
from mimeo.config.exc import UnsupportedPropertyValueError
from mimeo.config.mimeo_config import MimeoConfig
from mimeo.generators import Generator, JSONGenerator, XMLGenerator


class GeneratorFactory:
    """A Factory class instantiating a Generator based on Mimeo Config.

    Implementation of the Generator class depends on the output format
    configured.

    Attributes
    ----------
    XML
        The 'xml' output format
    JSON
        The 'json' output format

    Methods
    -------
    get_generator(mimeo_config: MimeoConfig) -> Generator
        Initialize a Generator based on the Mimeo Output Format.
    """

    XML: str = cc.OUTPUT_FORMAT_XML
    JSON: str = cc.OUTPUT_FORMAT_JSON

    @staticmethod
    def get_generator(
            mimeo_config: MimeoConfig,
    ) -> Generator:
        """Initialize a Generator based on the Mimeo Output Format.

        Parameters
        ----------
        mimeo_config : MimeoConfig
            A Mimeo Configuration

        Returns
        -------
        Generator
            A Generator's implementation instance

        Raises
        ------
        UnsupportedPropertyValueError
            If the output format is not supported
        """
        output_format = mimeo_config.output.format
        if output_format == GeneratorFactory.XML:
            return XMLGenerator(mimeo_config)
        if output_format == GeneratorFactory.JSON:
            return JSONGenerator(mimeo_config)
        raise UnsupportedPropertyValueError(
            cc.OUTPUT_FORMAT_KEY,
            output_format,
            cc.SUPPORTED_OUTPUT_FORMATS)



"""The Mimeo module.

This module is a main module providing the most high level class
in Mimeo:
    * Mimeograph
        A class responsible for the Mimeo processing.
"""
from __future__ import annotations

import asyncio
import logging
import os
import queue
import xml.etree.ElementTree as ElemTree
from concurrent.futures import ThreadPoolExecutor
from types import TracebackType
from typing import Iterable, Iterator

from mimeo.config.mimeo_config import MimeoConfig
from mimeo.consumers import ConsumerFactory
from mimeo.context import MimeoContextManager
from mimeo.exc import NotRunningMimeographError
from mimeo.generators import GeneratorFactory

logger = logging.getLogger(__name__)


class Mimeograph:
    """A class responsible for the Mimeo processing.

    Based on the Mimeo Configuration it instantiates generator and
    consumer to produce desired data.
    For sequential processing it is enough to use its static methods, without
    instantiation [process(), generate(), consume()]. However, Mimeograph exposes
    an interface to process data in parallel [run(), submit(), close()].

    Methods
    -------
    generate(
        mimeo_config: MimeoConfig,
        stringify: bool = False,
    ) -> Iterator[ElemTree.Element | dict | str]
        Generate data from the Mimeo Configuration.

    consume(
        mimeo_config: MimeoConfig,
        data: Iterable,
    )
        Consume data generated from the Mimeo Configuration.

    process(
        mimeo_config: MimeoConfig,
    )
        Process the Mimeo Configuration (generate data and consume).

    run(
    )
        Run the Mimeograph instance.
    stop(
    )
        Stop the Mimeograph instance.
    submit(
        mimeo_config: tuple[str | None, MimeoConfig | None],
    ):
        Put a Mimeo Config with identifier into a queue to process.

    Examples
    --------
    # Simple processing of a mimeo config
    config_path = "SomeEntity-config.json"
    mimeo_config = MimeoConfigFactory.parse(config_path)
    Mimeograph.process(mimeo_config)

    # Generating and consuming data from a mimeo config
    config_path = "SomeEntity-config.json"
    mimeo_config = MimeoConfigFactory.parse(config_path)
    data = Mimeograph.generate(mimeo_config)
    Mimeograph.consume(mimeo_config, data)

    # Processing mimeo configs in parallel
    config_paths = []
    with Mimeograph() as mimeo:
        for config_path in config_paths:
            mimeo_config = MimeoConfigFactory.parse(config_path)
            mimeo.submit((config_path, mimeo_config))

    # Processing mimeo configs in parallel outside context manager
    config_paths = []
    mimeo = Mimeograph()
    mimeo.run()
    for config_path in config_paths:
        mimeo_config = MimeoConfigFactory.parse(config_path)
        mimeo.submit((config_path, mimeo_config))
    mimeo.stop()
    """

    def __init__(
            self,
            workers: int = -1,
    ):
        """Initialize Mimeograph class.

        Parameters
        ----------
        workers : int
            A number of consumer workers
        """
        self._is_running: bool = False
        self._generator_queue: queue.Queue = queue.Queue()
        self._consumer_queue: queue.Queue = queue.Queue()
        self._generate_executor: ThreadPoolExecutor | None = None
        self._consume_executor: ThreadPoolExecutor | None = None
        self._failed_configs = []
        if workers == -1:
            self._consumer_workers = self._get_max_num_of_workers()
        else:
            self._consumer_workers = workers

    def __enter__(
            self,
    ) -> Mimeograph:
        """Enter the Mimeograph instance.

        It runs the Mimeograph instance.

        Returns
        -------
        self : Mimeograph
            A Mimeograph instance
        """
        self.run()
        return self

    def __exit__(
            self,
            exc_type: type | None,
            exc_val: BaseException | None,
            exc_tb: TracebackType | None,
    ) -> None:
        """Exit the Mimeograph instance.

        It stops the Mimeograph instance.

        Parameters
        ----------
        exc_type : type | None
            An exception's type
        exc_val : BaseException | None
            An exception's value
        exc_tb  TracebackType | None
            An exception's traceback
        """
        self.stop()

    def run(
            self,
    ):
        """Run the Mimeograph instance.

        It initializes generator and consumer workers and starts their tasks.
        """
        if not self._is_running:
            self._is_running = not self._is_running
            self._generate_executor = ThreadPoolExecutor(
                max_workers=1,
                thread_name_prefix="generator_thread")
            self._consume_executor = ThreadPoolExecutor(
                max_workers=self._consumer_workers,
                thread_name_prefix="consumer_thread")

            self._generate_executor.submit(self._start_generate)
            for _ in range(self._consumer_workers):
                self._consume_executor.submit(self._start_consume)

    def stop(
            self,
    ):
        """Stop the Mimeograph instance.

        It puts a poison pill to generator queue and awaits for all threads to stop
        before shutting executors down.
        """
        if self._is_running:
            self._is_running = not self._is_running
            self._generator_queue.put((None, None))
            self._generator_queue.join()
            self._consumer_queue.join()
            self._generate_executor.shutdown()
            self._consume_executor.shutdown()
            if len(self._failed_configs) > 0:
                logger.info("All configs have been processed. "
                            "The following configs have failed: %s",
                            self._failed_configs)
            else:
                logger.info("All configs have been successfully processed.")
            self._generate_executor = None
            self._consume_executor = None

    def submit(
            self,
            mimeo_config: tuple[str | None, MimeoConfig | None],
    ):
        """Put a Mimeo Config with identifier into a queue to process.

        Parameters
        ----------
        mimeo_config: tuple[str | None, MimeoConfig | None]
            A config ID and a Mimeo Config itself
        """
        if not self._is_running:
            raise NotRunningMimeographError
        logger.info("Putting a config for data generation into queue [%s]",
                    mimeo_config[0])
        self._generator_queue.put(mimeo_config)

    def _start_generate(
            self,
    ):
        """Start a generator task.

        Starts an infinitive loop that will work until a poison pill is being submitted.
        It gets a config from a queue and generates data. Once it took a poison pill,
        puts same to a consumer queue.
        """
        while True:
            logger.fine("Getting a config for data generation from queue")
            config_id, mimeo_config = self._generator_queue.get()
            if mimeo_config is None:
                self._stop_generate()
                break
            self._execute_generator_task(config_id, mimeo_config)

    def _execute_generator_task(
            self,
            config_id: str,
            mimeo_config: MimeoConfig,
    ):
        """Execute a generator task."""
        try:
            data = list(self.generate(mimeo_config, stringify=True))
            logger.fine("Putting data to consume to queue")
            self._consumer_queue.put((config_id, mimeo_config, data))
        except Exception:
            self._failed_configs.append(config_id)
            logger.exception("An unexpected error occurred while generating data "
                             "from a config [%s]", config_id)
        finally:
            self._generator_queue.task_done()

    def _stop_generate(
            self,
    ):
        """Stop a generator task."""
        logger.fine("Closing config generator")
        for _ in range(self._consumer_workers):
            self._consumer_queue.put((None, None, None))
        self._generator_queue.task_done()

    def _start_consume(
            self,
    ):
        """Start a consumer task.

        Starts an infinitive loop that will work until a poison pill is being submitted.
        It gets data from a queue and consumes it accordingly to a config.
        """
        while True:
            logger.fine("Getting data to consume from queue")
            config_id, mimeo_config, data = self._consumer_queue.get()
            if mimeo_config is None and data is None:
                self._stop_consume()
                break
            self._execute_consumer_task(config_id, mimeo_config, data)

    def _execute_consumer_task(
            self,
            config_id: str,
            mimeo_config: MimeoConfig,
            data: list,
    ):
        """Execute a consumer task."""
        try:
            self.consume(mimeo_config, data)
        except Exception:
            self._failed_configs.append(config_id)
            logger.exception("An unexpected error occurred while consuming data "
                             "from a config [%s]", config_id)
        finally:
            self._consumer_queue.task_done()

    def _stop_consume(
            self,
    ):
        """Stop a consumer task."""
        logger.fine("Closing data consumer")
        self._consumer_queue.task_done()

    @classmethod
    def process(
            cls,
            mimeo_config: MimeoConfig,
    ):
        """Process the Mimeo Configuration (generate data and consume).

        Parameters
        ----------
        mimeo_config: MimeoConfig
            A Mimeo Configuration to process
        """
        data = cls.generate(mimeo_config, stringify=True)
        cls.consume(mimeo_config, data)

        logger.info("Data has been processed")

    @classmethod
    def generate(
            cls,
            mimeo_config: MimeoConfig,
            stringify: bool = False,
    ) -> Iterator[ElemTree.Element | dict | str]:
        """Generate data from the Mimeo Configuration.

        Parameters
        ----------
        mimeo_config: MimeoConfig
            A Mimeo Configuration for data generation
        stringify: bool
            Indicate if data should be stringified

        Returns
        -------
        Iterator[ElemTree.Element | dict | str]
            Iterator for generated data
        """
        generator = GeneratorFactory.get_generator(mimeo_config)
        logger.info("Starting data generation")
        with MimeoContextManager(mimeo_config):
            for data in generator.generate(mimeo_config.templates):
                yield data if not stringify else generator.stringify(data)

    @classmethod
    def consume(
            cls,
            mimeo_config: MimeoConfig,
            data: Iterable,
    ):
        """Consume data generated from the Mimeo Configuration.

        Parameters
        ----------
        mimeo_config: MimeoConfig
            A Mimeo Configuration for data generation
        data: Iterable
            Data to consume
        """
        consumer = ConsumerFactory.get_consumer(mimeo_config)
        asyncio.run(consumer.consume(data))

    @staticmethod
    def _get_max_num_of_workers():
        """Get a maximum number of ThreadPoolExecutor workers."""
        return min(32, (os.cpu_count() or 1) + 4)  # Num of CPUs + 4



"""The Mimeo Logging Filters module.

This module contains all filters used in logging.
It exports the following filters:
    * RegularFilter
        A Filter subclass filtering out log records below INFO level.
    * DetailedFilter
        A Filter subclass filtering out log records above DEBUG level.
"""
from __future__ import annotations

from logging import DEBUG, INFO, Filter, LogRecord


class RegularFilter(Filter):
    """A Filter subclass filtering out log records below INFO level.

    It is meant to be used for logging info in regular format.

    Methods
    -------
    filter(record: LogRecord)
        Determine if the specified record is to be logged.
    """

    def filter(
            self,
            record: LogRecord,
    ) -> bool:
        """Determine if the specified record is to be logged.

        Parameters
        ----------
        record : LogRecord
            A log record

        Returns
        -------
        bool
            True if the log record's level is at least INFO
        """
        return record.levelno >= INFO


class DetailedFilter(Filter):
    """A Filter subclass filtering out log records above DEBUG level.

    It is meant to be used for logging info in detailed format.

    Methods
    -------
    filter(record: LogRecord)
        Determine if the specified record is to be logged.
    """

    def filter(
            self,
            record: LogRecord,
    ):
        """Determine if the specified record is to be logged.

        Parameters
        ----------
        record : LogRecord
            A log record

        Returns
        -------
        bool
            True if the log record's level is at most DEBUG
        """
        return record.levelno <= DEBUG



"""The Mimeo Raw Consumer module.

It exports only one class:
    * RawConsumer
        A Consumer implementation printing data in the standard output.
"""
from __future__ import annotations

from typing import Collection, Generator

from mimeo.consumers import Consumer


class RawConsumer(Consumer):
    """A Consumer implementation printing data in the standard output.

    This Consumer is instantiated for the 'stdout' output direction
    and prints data produced by Mimeo in the standard output.

    Methods
    -------
    consume
        Print data generated in the standard output.
    """

    async def consume(
            self,
            data: Collection | Generator,
    ) -> None:
        """Print data generated in the standard output.

        It is an implementation of Consumer's abstract method.

        Parameters
        ----------
        data : Collection | Generator
            Stringified data generated by Mimeo
        """
        for data_unit in data:
            print(data_unit)



"""The Cities module.

It exports classes related to cities CSV data:
    * City
        DTO class representing a single row in cities CSV data.
    * CitiesDB
        Class exposing READ operations on cities CSV data.
"""
from __future__ import annotations

from typing import ClassVar

import pandas
from pandas import DataFrame

from mimeo import tools
from mimeo.database.exc import InvalidIndexError


class City:
    """DTO class representing a single row in cities CSV data.

    Attributes
    ----------
    id : int
        An identifier
    name : str
        A city name
    name_ascii : str
        A city name in ASCII encoding
    country : str
        A country of a city
    """

    def __init__(
            self,
            identifier: str,
            name: str,
            name_ascii: str,
            country: str,
    ):
        """Initialize City class.

        Parameters
        ----------
        identifier : str
            An identifier
        name : str
            A city name
        name_ascii : str
            A city name in ASCII encoding
        country : str
            A country of a city
        """
        self.id: int = int(identifier)
        self.name: str = name
        self.name_ascii: str = name_ascii
        self.country: str = country

    def __str__(
            self,
    ) -> str:
        """Stringify the City instance.

        Returns
        -------
        str
            A stringified `dict` representation of the City instance
        """
        return str({
            "id": self.id,
            "name": self.name,
            "name_ascii": self.name_ascii,
            "country": self.country,
        })

    def __repr__(
            self,
    ) -> str:
        """Represent the City instance.

        Returns
        -------
        str
            A python representation of the City instance
        """
        return (f"City("
                f"id='{self.id}', "
                f"name='{self.name}', "
                f"name_ascii='{self.name_ascii}', "
                f"country='{self.country}')")


class CitiesDB:
    """Class exposing READ operations on cities CSV data.

    Attributes
    ----------
    NUM_OF_RECORDS : int
        A number of rows in cities CSV data

    Methods
    -------
    get_cities() -> list[City]
        Get all cities.
    get_cities_of(country_iso3: str) -> list[City]
        Get cities of a specific country.
    get_city_at(index: int) -> City
        Get a city at `index` position.
    """

    NUM_OF_RECORDS: int = 42904
    _CITIES_DB: str = "cities.csv"
    _CITIES_DF: DataFrame = None
    _CITIES: ClassVar[list] = None
    _COUNTRY_CITIES: ClassVar[dict] = {}

    def get_city_at(
            self,
            index: int,
    ) -> City:
        """Get a city at `index` position.

        Parameters
        ----------
        index : int
            A city row index

        Returns
        -------
        City
            A specific city

        Raises
        ------
        InvalidIndexError
            If the provided `index` is out of bounds
        """
        cities = CitiesDB._get_cities()
        try:
            return cities[index]
        except IndexError:
            last_index = CitiesDB.NUM_OF_RECORDS-1
            raise InvalidIndexError(index, last_index) from IndexError

    def get_cities_of(
            self,
            country_iso3: str,
    ) -> list[City]:
        """Get cities of a specific country.

        Parameters
        ----------
        country_iso3 : str
            A country ISO3 code to filter cities

        Returns
        -------
        list[City]
            List of cities filtered by country
        """
        return CitiesDB._get_country_cities(country_iso3).copy()

    def get_cities(
            self,
    ) -> list[City]:
        """Get all cities.

        Returns
        -------
        list[City]
            List of all cities
        """
        return CitiesDB._get_cities().copy()

    @classmethod
    def _get_country_cities(
            cls,
            country_iso3: str,
    ) -> list[City]:
        """Get cities of a specific country from cache.

        The country's cities list is initialized for the first time
        and cached in internal class attribute.

        Parameters
        ----------
        country_iso3 : str
            A country ISO3 code to filter cities

        Returns
        -------
        list[City]
            List of cities filtered by country
        """
        if country_iso3 not in cls._COUNTRY_CITIES:
            cities = cls._get_cities()
            cities = filter(lambda c: c.country == country_iso3, cities)
            cls._COUNTRY_CITIES[country_iso3] = list(cities)
        return cls._COUNTRY_CITIES[country_iso3]

    @classmethod
    def _get_cities(
            cls,
    ) -> list[City]:
        """Get all cities from cache.

        The cities list is initialized for the first time and cached
        in internal class attribute.

        Returns
        -------
        list[City]
            List of all cities
        """
        if cls._CITIES is None:
            cls._CITIES = [City(row.ID, row.CITY, row.CITY_ASCII, row.COUNTRY)
                           for row in cls._get_cities_df().itertuples()]
        return cls._CITIES

    @classmethod
    def _get_cities_df(
            cls,
    ) -> pandas.DataFrame:
        """Load cities CSV data and save in internal class attribute."""
        if cls._CITIES_DF is None:
            data = tools.get_resource(cls._CITIES_DB)
            cls._CITIES_DF = pandas.read_csv(data)
        return cls._CITIES_DF



"""The Last Names module.

It exports a class related to surnames CSV data:
    * LastNamesDB
        Class exposing READ operations on surnames CSV data.
"""
from __future__ import annotations

from typing import ClassVar

from mimeo import tools
from mimeo.database.exc import InvalidIndexError


class LastNamesDB:
    """Class exposing READ operations on surnames CSV data.

    Attributes
    ----------
    NUM_OF_RECORDS : int
        A number of rows in surnames CSV data

    Methods
    -------
    get_last_names() -> list[str]
        Get all last names.
    get_last_name_at(index: int) -> str
        Get a last name at `index` position.
    """

    NUM_OF_RECORDS: int = 151670
    _LAST_NAMES_DB: str = "surnames.txt"
    _LAST_NAMES: ClassVar[list] = None

    def get_last_name_at(
            self,
            index: int,
    ) -> str:
        """Get a last name at `index` position.

        Parameters
        ----------
        index : int
            A last name row index

        Returns
        -------
        str
            A last name

        Raises
        ------
        InvalidIndexError
            If the provided `index` is out of bounds
        """
        last_names = LastNamesDB._get_last_names()
        try:
            return last_names[index]
        except IndexError:
            last_index = LastNamesDB.NUM_OF_RECORDS-1
            raise InvalidIndexError(index, last_index) from IndexError

    def get_last_names(
            self,
    ) -> list[str]:
        """Get all last names.

        Returns
        -------
        list[str]
            List of all last names
        """
        return LastNamesDB._get_last_names().copy()

    @classmethod
    def _get_last_names(
            cls,
    ) -> list:
        """Get all last names from cache.

        The last names list is initialized for the first time and
        cached in internal class attribute.

        Returns
        -------
        list[str]
            List of all last names
        """
        if cls._LAST_NAMES is None:
            with tools.get_resource(cls._LAST_NAMES_DB) as last_names:
                cls._LAST_NAMES = [line.rstrip() for line in last_names.readlines()]
        return cls._LAST_NAMES
